from __future__ import unicode_literals
from __future__ import division
from babel.numbers import format_decimal
from babel.dates import format_datetime
from io import BytesIO
import copy
import datetime
import decimal
import PIL
import qrcode
import tempfile

from .barcode128 import code128_image
from .context import Context
from .docelement import DocElementBase, DocElement
from .enums import *
from .errors import Error, ReportBroError
from .rendering import ImageRenderElement, BarCodeRenderElement, TableRenderElement,\
    FrameRenderElement, SectionRenderElement
from .structs import Color, BorderStyle, TextStyle
from .utils import get_float_value, get_int_value, get_str_value, to_string, PY2, get_image_display_size

if PY2:
    import urllib2
else:
    import urllib

try:
    basestring  # For Python 2, str and unicode
except NameError:
    basestring = str


class ImageElement(DocElement):
    def __init__(self, report, data):
        DocElement.__init__(self, report, data)
        self.source = get_str_value(data, 'source')
        self.image = get_str_value(data, 'image')
        self.image_filename = get_str_value(data, 'imageFilename')
        self.horizontal_alignment = HorizontalAlignment[data.get('horizontalAlignment')]
        self.vertical_alignment = VerticalAlignment[data.get('verticalAlignment')]
        self.background_color = Color(data.get('backgroundColor'))
        self.print_if = get_str_value(data, 'printIf')
        self.remove_empty_element = bool(data.get('removeEmptyElement'))
        self.link = get_str_value(data, 'link')
        self.spreadsheet_hide = bool(data.get('spreadsheet_hide'))
        self.spreadsheet_column = get_int_value(data, 'spreadsheet_column')
        self.spreadsheet_add_empty_row = bool(data.get('spreadsheet_addEmptyRow'))
        self.image_key = None
        self.prepared_link = None

    def prepare(self, ctx, pdf_doc, only_verify):
        self.image_key = None
        # set image_key which is used to fetch cached images
        if self.source:
            if Context.is_parameter_name(self.source):
                # use current parameter value as image key
                param_ref = ctx.get_parameter(Context.strip_parameter_name(self.source))
                if param_ref:
                    source_parameter = param_ref.parameter
                    if source_parameter.type == ParameterType.string:
                        self.image_key, _ = ctx.get_parameter_data(param_ref)
                    elif source_parameter.type == ParameterType.image:
                        self.image_key = self.source + '_' +\
                                         str(Context.get_parameter_context_id(param_ref))
            else:
                # static url
                self.image_key = self.source
        else:
            # static image
            if self.image_filename:
                self.image_key = self.image_filename
            else:
                self.image_key = 'image_' + str(self.id)
        self.report.load_image(self.image_key, ctx, self.id, self.source, self.image)
        if self.link:
            self.prepared_link = ctx.fill_parameters(self.link, self.id, field='link')
            if not (self.prepared_link.startswith('http://') or self.prepared_link.startswith('https://')):
                raise ReportBroError(
                    Error('errorMsgInvalidLink', object_id=self.id, field='link'))

    def get_next_render_element(self, offset_y, container_top, container_height, ctx, pdf_doc):
        _, rv = DocElement.get_next_render_element(
            self, offset_y, container_top, container_height, ctx, pdf_doc)
        if not rv:
            return None, False
        return ImageRenderElement(self.report, offset_y, self), True

    def render_spreadsheet(self, row, col, ctx, renderer):
        if self.image_key:
            image = self.report.get_image(self.image_key)
            if self.spreadsheet_column:
                col = self.spreadsheet_column - 1

            try:
                raw_image = PIL.Image.open(image.image_fp)
            except Exception as ex:
                raise ReportBroError(
                    Error('errorMsgLoadingImageFailed', object_id=self.id,
                          field='source' if self.source else 'image', info=str(ex)))

            image_display_width, image_display_height = get_image_display_size(
                self.width, self.height, raw_image.width, raw_image.height)
            if image_display_width != raw_image.width or image_display_height != raw_image.height:
                raw_image = raw_image.resize(
                    (int(image_display_width), int(image_display_height)), PIL.Image.BILINEAR)
                image.image_fp = BytesIO()
                raw_image.save(image.image_fp, format='PNG' if image.image_type.upper() == 'PNG' else 'JPEG')

            renderer.insert_image(row, col, image_filename=self.image_filename, image_data=image.image_fp,
                                  width=self.width, url=self.prepared_link)
            row += 2 if self.spreadsheet_add_empty_row else 1
            col += 1
        return row, col


class BarCodeElement(DocElement):
    def __init__(self, report, data):
        DocElement.__init__(self, report, data)
        self.content = get_str_value(data, 'content')
        self.format = get_str_value(data, 'format').lower()
        assert self.format in ('code128', 'qrcode')
        self.display_value = bool(data.get('displayValue')) if self.format == 'code128' else False
        error_correction_level = get_str_value(data, 'errorCorrectionLevel')
        self.error_correction_level = qrcode.ERROR_CORRECT_M
        if error_correction_level == 'L':
            self.error_correction_level = qrcode.ERROR_CORRECT_L
        elif error_correction_level == 'H':
            self.error_correction_level = qrcode.ERROR_CORRECT_H
        elif error_correction_level == 'Q':
            self.error_correction_level = qrcode.ERROR_CORRECT_Q
        self.print_if = data.get('printIf', '')
        self.remove_empty_element = bool(data.get('removeEmptyElement'))
        self.spreadsheet_hide = bool(data.get('spreadsheet_hide'))
        self.spreadsheet_column = get_int_value(data, 'spreadsheet_column')
        self.spreadsheet_colspan = get_int_value(data, 'spreadsheet_colspan')
        self.spreadsheet_add_empty_row = bool(data.get('spreadsheet_addEmptyRow'))
        self.image_key = None
        self.image_height = self.height - 22 if (self.display_value and self.format == 'code128') else self.height
        self.prepared_content = None

    def is_printed(self, ctx):
        if not self.content:
            return False
        return DocElementBase.is_printed(self, ctx)

    def prepare(self, ctx, pdf_doc, only_verify):
        self.image_key = None
        self.prepared_content = ctx.fill_parameters(self.content, self.id, field='content')
        if self.prepared_content:
            try:
                if self.format == 'qrcode':
                    img = qrcode.make(self.prepared_content, border=0, error_correction=self.error_correction_level)
                elif self.format == 'code128':
                    img = code128_image(self.prepared_content, height=self.image_height, thickness=2, quiet_zone=False)
            except:
                raise ReportBroError(
                    Error('errorMsgInvalidBarCode', object_id=self.id, field='content'))
            if not only_verify:
                with tempfile.NamedTemporaryFile(delete=False, suffix='.png') as f:
                    img.save(f.name)
                    self.image_key = f.name
                    if self.format == 'qrcode':
                        # QR Code is always quadratic and only height can be set -> use same value for width
                        self.width = self.height
                    else:
                        # for Code128 the width is defined by the image width of the bar code
                        self.width = img.width

    def get_next_render_element(self, offset_y, container_top, container_height, ctx, pdf_doc):
        _, rv = DocElement.get_next_render_element(
            self, offset_y, container_top, container_height, ctx, pdf_doc)
        if not rv:
            return None, False
        return BarCodeRenderElement(self.report, offset_y, self), True

    def render_spreadsheet(self, row, col, ctx, renderer):
        if self.content:
            cell_format = dict()
            if self.spreadsheet_column:
                col = self.spreadsheet_column - 1
            renderer.write(row, col, self.spreadsheet_colspan, self.prepared_content, cell_format, self.width)
            row += 2 if self.spreadsheet_add_empty_row else 1
            col += 1
        return row, col


class LineElement(DocElement):
    def __init__(self, report, data):
        DocElement.__init__(self, report, data)
        self.color = Color(data.get('color'))
        self.print_if = get_str_value(data, 'printIf')

    def render_pdf(self, container_offset_x, container_offset_y, pdf_doc):
        pdf_doc.set_draw_color(self.color.r, self.color.g, self.color.b)
        pdf_doc.set_line_width(self.height)
        x = self.x + container_offset_x
        y = self.render_y + container_offset_y + (self.height / 2)
        pdf_doc.line(x, y, x + self.width, y)


class PageBreakElement(DocElementBase):
    def __init__(self, report, data):
        DocElementBase.__init__(self, report, data)
        self.id = get_int_value(data, 'id')
        self.print_if = get_str_value(data, 'printIf')
        self.x = 0
        self.width = 0
        self.sort_order = 0  # sort order for elements with same 'y'-value, render page break before other elements


class TextElement(DocElement):
    def __init__(self, report, data):
        DocElement.__init__(self, report, data)
        self.content = get_str_value(data, 'content')
        self.rich_text = bool(data.get('richText'))
        self.rich_text_content = data.get('richTextContent') or {}
        if self.rich_text and not isinstance(self.rich_text_content, dict):
            raise RuntimeError('Invalid richTextContent for text element {id}'.format(id=self.id))
        self.eval = bool(data.get('eval'))
        if data.get('styleId'):
            style = report.styles.get(get_int_value(data, 'styleId'))
            if style is None:
                raise RuntimeError('Style for text element {id} not found'.format(id=self.id))
            # shallow copy is sufficient in our case
            self.style = copy.copy(style)
        else:
            self.style = TextStyle(data)
        self.print_if = get_str_value(data, 'printIf')
        self.pattern = get_str_value(data, 'pattern')
        self.link = get_str_value(data, 'link')
        self.cs_condition = get_str_value(data, 'cs_condition')
        if self.cs_condition:
            if data.get('cs_styleId'):
                style = report.styles.get(int(data.get('cs_styleId')))
                if style is None:
                    raise RuntimeError('Conditional style for text element {id} not found'.format(id=self.id))
                # shallow copy is sufficient in our case
                self.conditional_style = copy.copy(style)
            else:
                self.conditional_style = TextStyle(data, key_prefix='cs_')
        else:
            self.conditional_style = None
        # additional styles are used when text is rendered inside table row and
        # the row has a background color -> a new style is created based on the
        # existing style
        self.additional_styles = dict()
        if isinstance(self, TableTextElement):
            self.remove_empty_element = False
            self.always_print_on_same_page = False
        else:
            self.remove_empty_element = bool(data.get('removeEmptyElement'))
            self.always_print_on_same_page = bool(data.get('alwaysPrintOnSamePage'))
        self.height = get_int_value(data, 'height')
        self.spreadsheet_hide = bool(data.get('spreadsheet_hide'))
        self.spreadsheet_column = get_int_value(data, 'spreadsheet_column')
        self.spreadsheet_colspan = get_int_value(data, 'spreadsheet_colspan')
        self.spreadsheet_add_empty_row = bool(data.get('spreadsheet_addEmptyRow'))
        self.spreadsheet_text_wrap = bool(data.get('spreadsheet_textWrap'))
        self.spreadsheet_formats = dict()  # caching of formats for rendering spreadsheet
        self.text_height = 0
        self.line_index = -1
        self.lines_count = 0
        self.text_lines = None
        self.used_style = None
        self.prepared_link = None
        self.space_top = 0
        self.space_bottom = 0
        self.total_height = 0
        self.spreadsheet_cell_format = None
        self.spreadsheet_cell_format_initialized = False

    def fill_parameters(self, ctx):
        return ctx.fill_parameters(self.content, self.id, field='content', pattern=self.pattern)

    def is_printed(self, ctx):
        if self.remove_empty_element and len(self.text_lines) == 0:
            return False
        return DocElementBase.is_printed(self, ctx)

    def prepare(self, ctx, pdf_doc, only_verify):
        if not self.rich_text:
            if self.eval:
                content = ctx.evaluate_expression(self.content, self.id, field='content')
                if self.pattern:
                    if isinstance(content, (int, float, decimal.Decimal)):
                        try:
                            content = format_decimal(content, self.pattern, locale=ctx.pattern_locale)
                            if self.pattern.find('$') != -1:
                                content = content.replace('$', ctx.pattern_currency_symbol)
                        except ValueError:
                            raise ReportBroError(
                                Error('errorMsgInvalidPattern', object_id=self.id, field='pattern', context=self.content))
                    elif isinstance(content, datetime.date):
                        try:
                            content = format_datetime(content, self.pattern, locale=ctx.pattern_locale)
                        except ValueError:
                            raise ReportBroError(
                                Error('errorMsgInvalidPattern', object_id=self.id, field='pattern', context=self.content))
                content = to_string(content)
            else:
                content = self.fill_parameters(ctx)

            if self.link:
                self.prepared_link = ctx.fill_parameters(self.link, self.id, field='link')
                if not (self.prepared_link.startswith('http://') or self.prepared_link.startswith('https://')):
                    raise ReportBroError(
                        Error('errorMsgInvalidLink', object_id=self.id, field='link'))
        else:
            content = None

        if self.cs_condition:
            if ctx.evaluate_expression(self.cs_condition, self.id, field='cs_condition'):
                self.used_style = self.conditional_style
            else:
                self.used_style = self.style
        else:
            self.used_style = self.style
        if self.used_style.vertical_alignment != VerticalAlignment.top and not self.always_print_on_same_page and\
                not isinstance(self, TableTextElement):
            self.always_print_on_same_page = True
        available_width = self.width - self.used_style.padding_left - self.used_style.padding_right

        self.text_lines = []
        if pdf_doc:
            self.split_text_lines(content, available_width=available_width, ctx=ctx, pdf_doc=pdf_doc)
            self.line_index = 0

            if isinstance(self, TableTextElement):
                self.total_height = max(
                    self.text_height + self.used_style.padding_top +
                    self.used_style.padding_bottom, self.height)
            else:
                self.set_height(self.height)
        else:
            # set text_lines so is_printed can check for empty element when rendering spreadsheet
            self.text_lines = [content] if content else []

    def get_style(self, style_id, background_color, base_style):
        if style_id in self.additional_styles:
            return self.additional_styles[style_id]
        # shallow copy is sufficient in our case
        style = copy.copy(base_style)
        style.id = style_id
        style.background_color = background_color
        self.additional_styles[style_id] = style
        return style

    def set_height(self, height):
        self.space_top = 0
        self.space_bottom = 0
        if self.text_height > 0:
            total_height = self.text_height + self.used_style.padding_top + self.used_style.padding_bottom
        else:
            total_height = 0
        if total_height < height:
            remaining_space = height - total_height
            if self.used_style.vertical_alignment == VerticalAlignment.top:
                self.space_bottom = remaining_space
            elif self.used_style.vertical_alignment == VerticalAlignment.middle:
                self.space_top = remaining_space / 2
                self.space_bottom = remaining_space / 2
            elif self.used_style.vertical_alignment == VerticalAlignment.bottom:
                self.space_top = remaining_space
        self.total_height = total_height + self.space_top + self.space_bottom

    def get_next_render_element(self, offset_y, container_top, container_height, ctx, pdf_doc):
        available_height = container_height - offset_y
        if self.always_print_on_same_page and self.first_render_element and\
                self.total_height > available_height and (offset_y != 0 or container_top != 0):
            return None, False

        lines = []
        remaining_height = available_height
        block_height = 0
        text_height = 0
        text_offset_y = 0
        if self.space_top > 0:
            space_top = min(self.space_top, remaining_height)
            self.space_top -= space_top
            block_height += space_top
            remaining_height -= space_top
            text_offset_y = space_top
        if self.space_top == 0:
            while self.line_index < self.lines_count:
                last_line = (self.line_index >= self.lines_count - 1)
                line_height = self.text_lines[self.line_index].height
                tmp_height = line_height
                if self.line_index == 0:
                    tmp_height += self.used_style.padding_top
                if  last_line:
                    tmp_height += self.used_style.padding_bottom
                if tmp_height > remaining_height:
                    break
                lines.append(self.text_lines[self.line_index])
                remaining_height -= tmp_height
                block_height += tmp_height
                text_height += line_height
                self.line_index += 1

        if self.line_index >= self.lines_count and self.space_bottom > 0:
            space_bottom = min(self.space_bottom, remaining_height)
            self.space_bottom -= space_bottom
            block_height += space_bottom
            remaining_height -= space_bottom

        if self.space_top == 0 and self.line_index == 0 and self.lines_count > 0:
            # even first line does not fit
            if offset_y != 0 or container_top != 0:
                # either container is not at top of page or element is not at top inside container
                # -> try on next page
                return None, False
            else:
                # already on top of container -> raise error
                raise ReportBroError(
                    Error('errorMsgInvalidSize', object_id=self.id, field='height'))

        rendering_complete = self.line_index >= self.lines_count and self.space_top == 0 and self.space_bottom == 0
        if not rendering_complete and remaining_height > 0:
            # draw text block until end of container
            block_height += remaining_height
            remaining_height = 0

        if self.first_render_element and rendering_complete:
            render_element_type = RenderElementType.complete
        else:
            if self.first_render_element:
                render_element_type = RenderElementType.first
            elif rendering_complete:
                render_element_type = RenderElementType.last
                if self.used_style.vertical_alignment == VerticalAlignment.bottom:
                    # make sure text is exactly aligned to bottom
                    tmp_offset_y = block_height - self.used_style.padding_bottom - text_height
                    if tmp_offset_y > 0:
                        text_offset_y = tmp_offset_y
            else:
                render_element_type = RenderElementType.between

        text_block_elem = TextBlockElement(
            self.report, x=self.x, y=self.y, render_y=offset_y,
            width=self.width, height=block_height, text_offset_y=text_offset_y,
            lines=lines, render_element_type=render_element_type, style=self.used_style)
        self.first_render_element = False
        self.render_bottom = text_block_elem.render_bottom
        self.rendering_complete = rendering_complete
        return text_block_elem, rendering_complete

    def is_first_render_element(self):
        return self.first_render_element

    def render_spreadsheet(self, row, col, ctx, renderer):
        cell_format = None
        if self.used_style.id not in self.spreadsheet_formats:
            format_props = dict()
            if self.used_style.bold:
                format_props['bold'] = True
            if self.used_style.italic:
                format_props['italic'] = True
            if self.used_style.underline:
                format_props['underline'] = True
            if self.used_style.strikethrough:
                format_props['font_strikeout'] = True
            if self.used_style.horizontal_alignment != HorizontalAlignment.left:
                format_props['align'] = self.used_style.horizontal_alignment.name
            if self.used_style.vertical_alignment != VerticalAlignment.top:
                if self.used_style.vertical_alignment == VerticalAlignment.middle:
                    format_props['valign'] = 'vcenter'
                else:
                    format_props['valign'] = self.used_style.vertical_alignment.name
            if not self.used_style.text_color.is_black():
                format_props['font_color'] = self.used_style.text_color.color_code
            if not self.used_style.background_color.transparent:
                format_props['bg_color'] = self.used_style.background_color.color_code
            if self.used_style.border_left or self.used_style.border_top or\
                    self.used_style.border_right or self.used_style.border_bottom:
                if not self.used_style.border_color.is_black():
                    format_props['border_color'] = self.used_style.border_color.color_code
                if self.used_style.border_left:
                    format_props['left'] = 1
                if self.used_style.border_top:
                    format_props['top'] = 1
                if self.used_style.border_right:
                    format_props['right'] = 1
                if self.used_style.border_bottom:
                    format_props['bottom'] = 1
            if self.spreadsheet_text_wrap:
                format_props['text_wrap'] = True
            if format_props:
                cell_format = renderer.add_format(format_props)
                self.spreadsheet_formats[self.used_style.id] = cell_format
        else:
            # use cached cell format which is already added to renderer
            cell_format = self.spreadsheet_formats[self.used_style.id]
        if self.spreadsheet_column:
            col = self.spreadsheet_column - 1
        content = self.get_spreadsheet_content()
        renderer.write(row, col, self.spreadsheet_colspan, content, cell_format,
                       self.width, url=self.prepared_link)
        if self.spreadsheet_add_empty_row:
            row += 1
        return row + 1, col + (self.spreadsheet_colspan if self.spreadsheet_colspan else 1)

    def get_spreadsheet_content(self):
        return self.text_lines[0] if self.text_lines else ''

    def set_font_by_style(self, style, pdf_doc):
        if not pdf_doc.set_font(
                style.font, style.font_style, style.font_size, underline=style.underline):
            error_field = style.key_prefix + 'font'
            raise ReportBroError(
                Error('errorMsgFontNotAvailable', object_id=self.id, field=error_field))

    def split_text_lines(self, content, available_width, ctx, pdf_doc):
        if content is not None:
            self.set_font_by_style(self.used_style, pdf_doc)
            try:
                lines = pdf_doc.split_text(first_w=available_width, w=available_width, txt=content)
            except UnicodeEncodeError:
                raise ReportBroError(
                    Error('errorMsgUnicodeEncodeError', object_id=self.id, field='content', context=self.content))
            for line in lines:
                text_line = TextLine(width=available_width, style=self.used_style, link=self.prepared_link)
                text, text_width, _ = line
                text_line.add_text(text, text_width, self.used_style)
                self.text_lines.append(text_line)
        else:
            if hasattr(self.__class__, 'split_rich_text_lines'):
                self.split_rich_text_lines(available_width, ctx, pdf_doc)
            else:
                raise ReportBroError(Error('errorMsgPlusVersionRequired', object_id=self.id, field='richText'))

        self.text_height = 0
        self.lines_count = len(self.text_lines)
        if self.lines_count > 0:
            self.text_lines[-1].last_line = True
            for text_line in self.text_lines:
                text_line.setup()
                self.text_height += text_line.height


class TextBlockElement(DocElementBase):
    def __init__(self, report, x, y, render_y, width, height, text_offset_y,
                 lines, render_element_type, style):
        DocElementBase.__init__(self, report, dict(y=y))
        self.x = x
        self.render_y = render_y
        self.render_bottom = render_y + height
        self.width = width
        self.height = height
        self.text_offset_y = text_offset_y
        self.lines = lines
        self.render_element_type = render_element_type
        self.style = style

    def render_pdf(self, container_offset_x, container_offset_y, pdf_doc):
        y = container_offset_y + self.render_y
        if not self.style.background_color.transparent:
            pdf_doc.set_fill_color(
                self.style.background_color.r, self.style.background_color.g, self.style.background_color.b)
            pdf_doc.rect(self.x + container_offset_x, y, self.width, self.height, style='F')
        if (self.style.border_left or self.style.border_top or
                self.style.border_right or self.style.border_bottom):
            DocElement.draw_border(
                x=self.x+container_offset_x, y=y, width=self.width, height=self.height,
                render_element_type=self.render_element_type, border_style=self.style, pdf_doc=pdf_doc)

        if self.render_element_type in (RenderElementType.complete, RenderElementType.first):
            y += self.style.padding_top
        y += self.text_offset_y

        if self.lines:
            self.lines[-1].last_line = True
            for line in self.lines:
                line.render_pdf(self.x + container_offset_x + self.style.padding_left, y, pdf_doc=pdf_doc)
                y += line.height


class TextLine(object):
    def __init__(self, width, style, link=None):
        self.available_width = width
        self.width = 0
        self.height = 0
        self.style = style
        self.link = link
        self.text = None
        self.last_line = False
        self.baseline_offset_y = 0

    def add_text(self, text, text_width, style, link=None):
        self.text = TextLinePart(text, text_width, style, link)

    def setup(self):
        self.width = self.text.width
        max_font_size = self.style.font_size
        self.baseline_offset_y = max_font_size * 0.8
        self.height = max_font_size if self.last_line else (max_font_size * self.style.line_spacing)

    def render_pdf(self, x, y, pdf_doc):
        offset_x = 0
        # last line of justified text is aligned left
        if self.style.horizontal_alignment != HorizontalAlignment.justify or self.last_line:
            if self.style.horizontal_alignment == HorizontalAlignment.center:
                offset_x = ((self.available_width - self.width) / 2)
            elif self.style.horizontal_alignment == HorizontalAlignment.right:
                offset_x = self.available_width - self.width
            render_x = x + offset_x
            render_y = y + self.baseline_offset_y

            pdf_doc.set_font(family=self.style.font, style=self.style.font_style,
                             size=self.style.font_size, underline=self.style.underline)
            pdf_doc.set_text_color(self.style.text_color.r, self.style.text_color.g, self.style.text_color.b)
            pdf_doc.text(render_x, render_y, self.text.text)
            if self.style.strikethrough:
                # use underline thickness
                strikethrough_thickness = pdf_doc.current_font['ut']
                render_line_y = render_y - self.style.font_size * 0.3
                strikethrough_width = strikethrough_thickness / 1000.0 * self.style.font_size
                pdf_doc.set_line_width(strikethrough_width)
                pdf_doc.set_draw_color(self.style.text_color.r, self.style.text_color.g, self.style.text_color.b)
                pdf_doc.line(render_x, render_line_y, render_x + self.text.width, render_line_y)
            if self.text.link:
                pdf_doc.link(render_x, y, self.text.width, self.style.font_size, self.text.link)
            render_x += self.text.width
        else:
            # justify text -> split text into words and use equal space between words
            words = []
            total_word_width = 0
            font_size = self.style.font_size

            pdf_doc.set_font(family=self.style.font, style=self.style.font_style,
                             size=font_size, underline=self.style.underline)
            pwords = self.text.text.split()
            for pword in pwords:
                tmp_width = pdf_doc.get_string_width(pword)
                total_word_width += tmp_width
                words.append((pword, tmp_width))

            word_count = len(words)
            spaces_count = word_count - 1
            word_spacing = ((self.available_width - total_word_width) / spaces_count) if spaces_count > 0 else 0
            render_y = y + self.baseline_offset_y

            # draw words with equal space between words
            pdf_doc.set_text_color(self.style.text_color.r, self.style.text_color.g, self.style.text_color.b)
            word_x = x
            for text, word_width in words:
                pdf_doc.text(word_x, render_y, text)
                word_x += word_width + word_spacing

            if self.style.strikethrough:
                line_position = pdf_doc.current_font['up']
                line_thickness = pdf_doc.current_font['ut']
                # strikethrough with an offset from baseline
                render_y = y + self.baseline_offset_y - font_size * 0.3
                line_width = line_thickness / 1000.0 * font_size
                pdf_doc.set_line_width(line_width)
                pdf_doc.set_draw_color(self.style.text_color.r, self.style.text_color.g, self.style.text_color.b)
                pdf_doc.line(x, render_y, x + self.available_width, render_y)

            if self.style.underline:
                line_position = pdf_doc.current_font['up']
                line_thickness = pdf_doc.current_font['ut']
                # underline needs offset for line position
                render_y = y + self.baseline_offset_y - (line_position / 1000.0 * font_size)
                line_width = line_thickness / 1000.0 * font_size
                pdf_doc.set_line_width(line_width)
                pdf_doc.set_draw_color(self.style.text_color.r, self.style.text_color.g, self.style.text_color.b)
                pdf_doc.line(x, render_y, x + self.available_width, render_y)

        if self.link:
            pdf_doc.link(x + offset_x, y, self.width, self.style.font_size, self.link)


class TextLinePart:
    def __init__(self, text, width, style, link):
        self.text = text
        self.width = width
        self.style = style
        self.link = link


class TableTextElement(TextElement):
    def __init__(self, report, data, simple_array_param=None, simple_array_item_index=-1):
        TextElement.__init__(self, report, data)
        self.colspan = get_int_value(data, 'colspan') if data.get('colspan') else 1
        # overwrite spreadsheet colspan with table cell colspan (spreadsheet colspan
        # cannot be set separately in a table cell)
        self.spreadsheet_colspan = self.colspan
        self.initial_width = self.width
        self.grow_weight = get_int_value(data, 'growWeight')
        # column can be hidden in case print_if condition is set in header and evaluates to False
        self.column_visible = True
        # a previous cell can have a colspan which includes this cell -> this cell is not shown
        self.inside_colspan = False
        self.simple_array_param = simple_array_param
        self.simple_array_item_index = simple_array_item_index
        self.data = data  # needed in case cell is expanded by simple_array parameter

    def expand_simple_array(self, printed_cells, ctx):
        if self.content and not self.eval and Context.is_parameter_name(self.content):
            param_ref = ctx.get_parameter(Context.strip_parameter_name(self.content))
            if param_ref:
                column_data_parameter = param_ref.parameter
                if column_data_parameter.type == ParameterType.simple_array:
                    cell_values, value_exists = ctx.get_parameter_data(param_ref)
                    if value_exists and cell_values and len(cell_values) > 0:
                        self.simple_array_param = column_data_parameter
                        self.simple_array_item_index = 0
                        for i in range(1, len(cell_values)):
                            expanded_cell = TableTextElement(
                                self.report, self.data,
                                simple_array_param=column_data_parameter, simple_array_item_index=i)
                            printed_cells.append(expanded_cell)

    def fill_parameters(self, ctx):
        if self.simple_array_param is not None:
            param_ref = ctx.get_parameter(self.simple_array_param.name)
            if param_ref:
                cell_values, value_exists = ctx.get_parameter_data(param_ref)
                if value_exists and self.simple_array_item_index < len(cell_values):
                    simple_array_item_value = cell_values[self.simple_array_item_index]
                    return ctx.get_formatted_value(
                        simple_array_item_value, self.simple_array_param,
                        object_id=None, is_array_item=True)
        return TextElement.fill_parameters(self, ctx)

    def is_printed(self, ctx):
        return self.column_visible and not self.inside_colspan


class TableElement(DocElement):
    def __init__(self, report, data):
        DocElement.__init__(self, report, data)
        self.data_source = get_str_value(data, 'dataSource')
        self.columns = get_int_value(data, 'columns')
        self.header = None
        self.content_rows = []
        self.row_index_after_main_content = -1
        self.has_table_band_group = False
        self.footer = None
        self.data_source_parameter = None

        column_count = None
        if bool(data.get('header')):
            self.header = TableBandElement(report, data.get('headerData'), BandType.header)
            column_count = len(self.header.cells)
        content_data_rows = data.get('contentDataRows')
        assert isinstance(content_data_rows, list)
        main_content_created = False
        for idx, content_data_row in enumerate(content_data_rows):
            band_element = TableBandElement(
                report, content_data_row, BandType.content, before_group=not main_content_created)

            if band_element.group_expression:
                self.has_table_band_group = True
            if main_content_created:
                if self.row_index_after_main_content == -1 and band_element.group_expression:
                    self.row_index_after_main_content = idx
            else:
                if not band_element.group_expression:
                    main_content_created = True

            self.content_rows.append(band_element)
            if column_count is None:
                column_count = len(band_element.cells)
            else:
                assert column_count == len(band_element.cells)
        if bool(data.get('footer')):
            self.footer = TableBandElement(report, data.get('footerData'), BandType.footer)
            if column_count is not None:
                assert column_count == len(self.footer.cells)

        self.print_header = False
        self.print_footer = False
        self.border = Border[data.get('border')]
        self.border_color = Color(data.get('borderColor'))
        self.border_width = get_float_value(data, 'borderWidth')
        self.print_if = get_str_value(data, 'printIf')
        self.remove_empty_element = bool(data.get('removeEmptyElement'))
        self.spreadsheet_hide = bool(data.get('spreadsheet_hide'))
        self.spreadsheet_column = get_int_value(data, 'spreadsheet_column')
        self.spreadsheet_add_empty_row = bool(data.get('spreadsheet_addEmptyRow'))
        self.row_parameters = dict()
        self.rows = []
        self.row_count = 0
        self.row_index = -1
        self.content_row_index = -1
        self.width = 0  # width will be set in prepare method
        if self.header:
            self.height += self.header.height
        if self.footer:
            self.height += self.footer.height
        if len(self.content_rows) > 0:
            for content_row in self.content_rows:
                self.height += content_row.height
        self.bottom = self.y + self.height

    def prepare(self, ctx, pdf_doc, only_verify):
        if self.header:
            self.print_header = True
            free_space = 0  # space freed up by hidden columns
            total_weight = 0
            for column_idx, cell in enumerate(self.header.cells):
                if not cell.inside_colspan and cell.print_if:
                    cell.column_visible = ctx.evaluate_expression(
                        cell.print_if, cell.id, field='printIf')
                    if not cell.column_visible:
                        free_space += cell.width
                    for content_row in self.content_rows:
                        content_row.cells[column_idx].column_visible = cell.column_visible
                    if self.footer:
                        self.footer.cells[column_idx].column_visible = cell.column_visible
                if cell.column_visible:
                    total_weight += cell.grow_weight

            # there is free space (because of hidden columns) and growable columns exist
            # -> the free space is shared among the growable columns depending on
            # their grow weight
            if free_space > 0 and total_weight > 0:
                # convert to float so division result is also float in Python 2
                total_weight = float(total_weight)
                for column_idx, cell in enumerate(self.header.cells):
                    if cell.grow_weight > 0:
                        added_width = int((free_space / total_weight) * cell.grow_weight + 0.5)
                        cell.width = cell.initial_width + added_width
                        for content_row in self.content_rows:
                            content_row.cells[column_idx].width = cell.width
                        if self.footer:
                            self.footer.cells[column_idx].width = cell.width

        if self.footer:
            self.print_footer = True

        parameter_name = Context.strip_parameter_name(self.data_source)
        if parameter_name:
            param_ref = ctx.get_parameter(parameter_name)
            if param_ref is None:
                raise ReportBroError(
                    Error('errorMsgMissingParameter', object_id=self.id, field='dataSource'))
            self.data_source_parameter = param_ref.parameter
            if self.data_source_parameter.type != ParameterType.array:
                raise ReportBroError(
                    Error('errorMsgInvalidDataSourceParameter', object_id=self.id, field='dataSource'))
            for row_parameter in self.data_source_parameter.children:
                self.row_parameters[row_parameter.name] = row_parameter
            self.rows, parameter_exists = ctx.get_parameter_data(param_ref)
            if not parameter_exists:
                raise ReportBroError(
                    Error('errorMsgMissingData', object_id=self.id, field='dataSource'))
            if not isinstance(self.rows, list):
                raise ReportBroError(
                    Error('errorMsgInvalidDataSource', object_id=self.id, field='dataSource'))
        else:
            # there is no data source parameter so we create a static table (faked by one empty data row)
            self.rows = [dict()]
            self.data_source_parameter = None

        self.row_count = len(self.rows)
        self.row_index = 0
        self.content_row_index = 0

        self.width = 0
        table_width_initialized = False
        # expand cells if necessary (a cell with a simple_array parameter
        # will be expanded to multiple cells) for all table bands and set table width
        if self.header:
            self.header.set_printed_cells(ctx)
            self.header.prepare(ctx)
            for cell in self.header.printed_cells:
                self.width += cell.width
            table_width_initialized = True

        if self.row_index < self.row_count:
            ctx.push_context(self.row_parameters, self.rows[self.row_index])
            for content_row in self.content_rows:
                content_row.set_printed_cells(ctx)
                if not table_width_initialized and not content_row.group_expression:
                    for cell in content_row.printed_cells:
                        self.width += cell.width
                    table_width_initialized = True
            ctx.pop_context()

        self.init_group_rows(ctx=ctx)

        if only_verify:
            # call prepare for each content band in each row to verify
            # group and print-if expressions
            self.row_index = 0
            while self.row_index < self.row_count:
                # push data context of current row so values of current row can be accessed
                ctx.push_context(self.row_parameters, self.rows[self.row_index])
                for content_row in self.content_rows:
                    content_row.prepare(ctx=ctx, row_index=self.row_index)
                ctx.pop_context()
                self.row_index += 1

        if self.footer:
            self.footer.set_printed_cells(ctx)
            self.footer.prepare(ctx)

    def get_next_render_element(self, offset_y, container_top, container_height, ctx, pdf_doc):
        self.render_y = offset_y
        self.render_bottom = self.render_y
        if self.is_rendering_complete():
            self.rendering_complete = True
            return None, True
        render_element = TableRenderElement(self.report, table=self, render_y=offset_y)

        if self.print_header:
            if not self.header.rendering_complete:
                self.header.create_render_elements(offset_y, container_top, container_height, ctx, pdf_doc)
            render_element.add_band(self.header)
            if not self.header.rendering_complete:
                return render_element, False
            if not self.header.repeat_header:
                self.print_header = False

        first_render_row_index = self.row_index
        first_content_row_index = self.content_row_index
        first_row_on_page = True
        while self.row_index < self.row_count:
            # push data context of current row so values of current row can be accessed
            ctx.push_context(self.row_parameters, self.rows[self.row_index])
            self.content_row_index = 0
            for content_row in self.content_rows:
                # in case we are on the first row of the page we have to check if the content row
                # is visible (and not already displayed on previous page) or if it is a group content
                # row which is repeated on every page
                if not first_row_on_page or self.content_row_index >= first_content_row_index or\
                        content_row.repeat_group_header:
                    content_row.prepare(ctx=ctx, row_index=self.row_index)
                    if content_row.repeat_group_header and first_row_on_page:
                        # group content row is repeated on every page
                        content_row.group_changed = True

                    if content_row.is_printed(ctx=ctx):
                        # only perform page break before content if there is at least
                        # one rendered row
                        if content_row.page_break and content_row.before_group and\
                                self.row_index != first_render_row_index:
                            ctx.pop_context()
                            return render_element, False

                        content_row.set_parameter_range(parameter=self.data_source_parameter, ctx=ctx)

                        content_row.create_render_elements(
                            offset_y + render_element.height, container_top, container_height, ctx, pdf_doc)

                        content_row.set_parameter_range(parameter=self.data_source_parameter, ctx=ctx, clear=True)

                        render_element.add_band(content_row, row_index=self.row_index)
                        if not content_row.rendering_complete:
                            ctx.pop_context()
                            return render_element, False

                        # only perform page break after content if this is not the last row
                        if content_row.page_break and not content_row.before_group and\
                                self.row_index < (self.row_count - 1):
                            self.content_row_index += 1
                            ctx.pop_context()
                            return render_element, False
                    else:
                        content_row.rendering_complete = True
                self.content_row_index += 1
            ctx.pop_context()

            self.row_index += 1
            first_row_on_page = False

        if self.row_index >= self.row_count and self.print_footer:
            self.footer.create_render_elements(
                offset_y + render_element.height, container_top, container_height, ctx, pdf_doc)
            render_element.add_band(self.footer)
            if not self.footer.rendering_complete:
                return render_element, False
            self.print_footer = False

        if self.is_rendering_complete():
            self.rendering_complete = True

        if render_element.is_empty():
            return None, self.rendering_complete

        self.render_bottom = render_element.render_bottom
        return render_element, self.rendering_complete

    def init_group_rows(self, ctx):
        """
        Initialize table group bands (e.g. to show header or sums for grouped rows) and
        set row index when a group band is printed.

        :param ctx: current context
        """
        if self.has_table_band_group and self.row_count > 0:
            row_index = 0
            content_group_rows = []
            for content_row in self.content_rows:
                if content_row.group_expression:
                    content_row.group_changed_row_indices = []
                    content_group_rows.append(content_row)

            # set next group expr which is used for group expr of first row
            ctx.push_context(self.row_parameters, self.rows[row_index])
            for content_group_row in content_group_rows:
                content_group_row.set_next_group_expression(ctx)
            ctx.pop_context()

            while row_index < self.row_count:
                has_next = (row_index + 1) < self.row_count
                # set context for next row (if available) so group expr of next row can be evaluated
                if has_next:
                    ctx.push_context(self.row_parameters, self.rows[row_index + 1])
                for content_group_row in content_group_rows:
                    # set group expression for next row so the group expressions for previous, next and current
                    # row are available
                    content_group_row.set_next_group_expression(ctx, has_next=has_next)
                    # now test if the group expression (prev <-> current for group before content or
                    # current <-> next for group after content has changed)
                    content_group_row.test_group_changed(row_index)
                if has_next:
                    ctx.pop_context()
                row_index += 1

    def is_rendering_complete(self):
        # test if footer and all content rows are completely rendered
        if self.row_index >= self.row_count and not self.print_footer:
            # only test rendering_complete of bands if at least one row was processed
            # because otherwise the flag was not set
            if self.row_index > 0:
                for content_row in self.content_rows:
                    if not content_row.rendering_complete:
                        return False
            return True
        else:
            return False

    def render_spreadsheet(self, row, col, ctx, renderer):
        if self.spreadsheet_column:
            col = self.spreadsheet_column - 1
        columns = 0

        if self.header:
            columns = len(self.header.printed_cells)
            row, _ = self.header.render_spreadsheet(row, col, ctx, renderer)

        while self.row_index < self.row_count:
            # push data context of current row so values of current row can be accessed
            ctx.push_context(self.row_parameters, self.rows[self.row_index])
            for i, content_row in enumerate(self.content_rows):
                content_row.prepare(ctx=ctx, row_index=self.row_index)
                if content_row.is_printed(ctx=ctx):
                    if columns == 0:
                        # get column count from first printed content row if there is no header
                        columns = len(content_row.printed_cells)
                    row, _ = content_row.render_spreadsheet(
                        row, col, ctx, renderer, row_index=self.row_index)
            ctx.pop_context()
            self.row_index += 1

        if self.footer:
            row, _ = self.footer.render_spreadsheet(row, col, ctx, renderer)

        if self.spreadsheet_add_empty_row:
            row += 1
        return row, col + columns


class TableBandElement(object):
    def __init__(self, report, data, band_type, before_group=False):
        from .containers import Container
        assert(isinstance(data, dict))
        self.id = get_int_value(data, 'id')
        self.height = get_int_value(data, 'height')
        self.band_type = band_type
        if band_type == BandType.header:
            self.repeat_header = bool(data.get('repeatHeader'))
        else:
            self.repeat_header = None
        self.background_color = Color(data.get('backgroundColor'))
        self.group_expression = get_str_value(data, 'groupExpression')
        self.print_if = get_str_value(data, 'printIf')
        self.before_group = before_group
        self.page_break = False
        self.repeat_group_header = None
        if band_type == BandType.content:
            self.alternate_background_color = Color(data.get('alternateBackgroundColor'))
            self.always_print_on_same_page = bool(data.get('alwaysPrintOnSamePage'))
            if self.group_expression:
                self.page_break = bool(data.get('pageBreak'))
                self.repeat_group_header = bool(data.get('repeatGroupHeader'))
                if self.repeat_group_header and not before_group:
                    report.errors.append(Error(
                        'errorMsgRepeatGroupHeaderAfterContent', object_id=self.id, field='repeatGroupHeader'))
        else:
            self.alternate_background_color = None
            self.always_print_on_same_page = True
        self.cells = []  # cells created from initial data as defined in Designer
        self.print_if_result = True

        self.group_changed = False
        self.group_expr_result = None
        self.prev_group_expr_result = None
        self.next_group_expr_result = None
        self.group_changed_row_indices = []
        self.group_row_index_start = 0
        self.group_row_index_end = 0

        # cells which will be printed, this excludes cells within a colspan of another cells
        # and can include additional cells which get expanded by a cell
        # with a simple_list parameter
        self.printed_cells = []

        colspan_end_idx = 0
        colspan_element = None
        column_data = data.get('columnData')
        assert isinstance(column_data, list)
        for idx, column in enumerate(column_data):
            assert isinstance(column, dict)
            column['height'] = self.height  # set height of cell to band height
            cell = TableTextElement(report, column)
            if idx < colspan_end_idx:
                colspan_element.initial_width += cell.width
                colspan_element.width += cell.width
                cell.inside_colspan = True
            elif cell.colspan > 1:
                colspan_element = cell
                colspan_end_idx = idx + cell.colspan
            self.cells.append(cell)

        # create a virtual container for each table band
        self.container = Container(
            container_id='tablerow_' + str(self.id), containers=None, report=report)
        self.container.height = self.height
        self.container.allow_page_break = False
        self.rendering_complete = False
        self.prepare_container = True
        self.rendered_band_height = 0

    def set_printed_cells(self, ctx):
        """Initialize the printed cells.

        Cells can be expanded by a simple_list parameter into multiple cells. Cells can also
        be hidden be other cells with colspan set. Must be called exactly once for this band.

        :param ctx: current context
        """
        printed_cells = []
        for cell in self.cells:
            if cell.column_visible and not cell.inside_colspan:
                printed_cells.append(cell)
                cell.expand_simple_array(printed_cells, ctx)
        self.printed_cells = printed_cells
        table_width = 0
        for cell in printed_cells:
            cell.x = table_width
            table_width += cell.width
            self.container.add(cell)
        self.container.width = table_width

    def set_next_group_expression(self, ctx, has_next=True):
        """Evaluate group expression of next row and set group expression of
        previous and current row accordingly.

        :param ctx: context where parameters of next row are pushed.
        :param has_next: True if next row exists.
        """
        self.prev_group_expr_result = self.group_expr_result
        self.group_expr_result = self.next_group_expr_result
        if has_next:
            self.next_group_expr_result = ctx.evaluate_expression(
                self.group_expression, self.id, field='groupExpression')
        else:
            self.next_group_expr_result = None

    def test_group_changed(self, row_index):
        """
        Test if group expression changed and store the row_index if this is the case.

        :param row_index: current row index, group expression for previous, current and next row must
        be set before with :func:`~TableBandElement.set_next_group_expression`
        """
        if self.group_expression:
            if self.before_group:
                if self.group_expr_result != self.prev_group_expr_result:
                    self.group_changed_row_indices.append(row_index)
            else:
                if self.group_expr_result != self.next_group_expr_result:
                    self.group_changed_row_indices.append(row_index)

    def prepare(self, ctx, row_index=None):
        if self.group_expression:
            self.group_changed = False
            if self.group_changed_row_indices and self.group_changed_row_indices[0] == row_index:
                # the group expression has changed for this row index -> group band will be printed
                self.group_changed = True
                # now set index range which is used for function parameters like sum/avg
                if self.before_group:
                    self.group_row_index_start = self.group_changed_row_indices[0]
                    if len(self.group_changed_row_indices) > 1:
                        self.group_row_index_end = self.group_changed_row_indices[1]
                    else:
                        self.group_row_index_end = -1
                else:
                    self.group_row_index_start = self.group_row_index_end
                    self.group_row_index_end = self.group_changed_row_indices[0] + 1

        if self.print_if:
            self.print_if_result = ctx.evaluate_expression(
                self.print_if, self.id, field='printIf')

    def set_parameter_range(self, parameter, ctx, clear=False):
        """
        Set start and end row index for the given parameter which is used when evaluating a parameter
        function (e.g. sum or avg).

        :param parameter: data source parameter of table
        :param ctx: current context
        :param clear: if True the range will be removed for the parameter
        """
        if self.group_expression:
            if clear:
                parameter.clear_range()
                ctx.dec_range_count()
            else:
                parameter.set_range(self.group_row_index_start, self.group_row_index_end)
                ctx.inc_range_count()

    def is_printed(self, ctx):
        return self.print_if_result and (not self.group_expression or self.group_changed)

    def render_spreadsheet(self, row, col, ctx, renderer, row_index=-1):
        """Render table band in spreadsheet.

        This takes care of the background color and sets it for the row if there is a color.
        The spreadsheet is rendered by calling the method for every element in
        the table band container.
        """

        # elements in container must be prepared for each row before spreadsheet can be rendered
        self.container.prepare(ctx, pdf_doc=None)

        background_color = self.background_color
        if self.band_type == BandType.content and not self.alternate_background_color.transparent and\
                row_index % 2 == 1:
            background_color = self.alternate_background_color
            style_id_suffix = '_alt_table_row'
        else:
            style_id_suffix = '_table_row'

        if not background_color.transparent:
            for cell in self.container.doc_elements:
                if isinstance(cell, TableTextElement):
                    base_style = cell.used_style
                    cell.used_style = cell.get_style(
                        base_style.id + style_id_suffix, background_color, base_style)

        return self.container.render_spreadsheet(row, col, ctx, renderer)

    def create_render_elements(self, offset_y, container_top, container_height, ctx, pdf_doc):
        available_height = container_height - offset_y
        if self.always_print_on_same_page and available_height < self.height:
            # not enough space for whole band
            self.rendering_complete = False
        else:
            if self.prepare_container:
                self.container.prepare(ctx, pdf_doc)
                self.rendered_band_height = 0

                heights = [self.height]
                # get max height of all cells of this band
                for cell in self.container.doc_elements:
                    if isinstance(cell, TableTextElement):
                        heights.append(cell.total_height)
                # all cells will be set to max cell height
                max_height = max(heights)
                for cell in self.container.doc_elements:
                    if isinstance(cell, TableTextElement):
                        cell.set_height(max_height)
            else:
                self.rendered_band_height += self.container.used_band_height
                # clear render elements from previous page
                self.container.clear_rendered_elements()

            self.rendering_complete = self.container.create_render_elements(
                container_top + offset_y, available_height, ctx=ctx, pdf_doc=pdf_doc)

        if self.rendering_complete:
            remaining_min_height = self.height - self.rendered_band_height
            if self.container.used_band_height < remaining_min_height:
                # rendering of band complete, make sure band is at least as large
                # as minimum height (even if it spans over more than 1 page)
                if remaining_min_height <= available_height:
                    self.prepare_container = True
                    # TODO: check
                    self.container.used_band_height = remaining_min_height
                else:
                    # minimum height is larger than available space, continue on next page
                    self.rendering_complete = False
                    self.prepare_container = False
                    self.container.used_band_height = available_height
            else:
                self.prepare_container = True
        else:
            if self.always_print_on_same_page:
                # band must be printed on same page but available space is not enough,
                # try to render it on top of next page
                self.prepare_container = True
                if offset_y == 0:
                    field = 'alwaysPrintOnSamePage' if self.band_type == BandType.content else 'size'
                    raise ReportBroError(
                        Error('errorMsgSectionBandNotOnSamePage', object_id=self.id, field=field))
            else:
                self.prepare_container = False
                self.container.used_band_height = available_height

        if self.rendering_complete and self.group_changed:
            self.group_changed_row_indices.pop(0)

    def get_used_band_height(self):
        return self.container.used_band_height

    def get_render_elements(self):
        return self.container.render_elements


class FrameElement(DocElement):
    def __init__(self, report, data, containers):
        DocElement.__init__(self, report, data)
        from .containers import Frame
        self.background_color = Color(data.get('backgroundColor'))
        self.border_style = BorderStyle(data)
        self.print_if = get_str_value(data, 'printIf')
        self.remove_empty_element = bool(data.get('removeEmptyElement'))
        self.shrink_to_content_height = bool(data.get('shrinkToContentHeight'))
        self.spreadsheet_hide = bool(data.get('spreadsheet_hide'))
        self.spreadsheet_column = get_int_value(data, 'spreadsheet_column')
        self.spreadsheet_add_empty_row = bool(data.get('spreadsheet_addEmptyRow'))

        # rendering_complete status for next page, in case rendering was not started on first page.
        self.next_page_rendering_complete = False
        # container content height of previous page, in case rendering was not started on first page
        self.prev_page_content_height = 0

        self.render_element_type = RenderElementType.none
        self.container = Frame(
            width=self.width, height=self.height,
            container_id=str(data.get('linkedContainerId')), containers=containers, report=report)

    def get_used_height(self):
        height = self.container.get_render_elements_bottom()
        if self.border_style.border_top and self.render_element_type == RenderElementType.none:
            height += self.border_style.border_width
        if self.border_style.border_bottom:
            height += self.border_style.border_width
        if self.render_element_type == RenderElementType.none and not self.shrink_to_content_height:
            height = max(self.height, height)
        return height

    def prepare(self, ctx, pdf_doc, only_verify):
        self.container.prepare(ctx, pdf_doc=pdf_doc, only_verify=only_verify)
        self.next_page_rendering_complete = False
        self.prev_page_content_height = 0
        self.render_element_type = RenderElementType.none

    def get_next_render_element(self, offset_y, container_top, container_height, ctx, pdf_doc):
        self.render_y = offset_y
        content_height = container_height
        render_element = FrameRenderElement(self.report, self, render_y=offset_y)

        if self.border_style.border_top and self.render_element_type == RenderElementType.none:
            content_height -= self.border_style.border_width
        if self.border_style.border_bottom:
            # this is not 100% correct because bottom border is only applied if frame fits
            # on current page. this should be negligible because the border is usually only a few pixels
            # and most of the time the frame fits on one page.
            # to get the exact height in advance would be quite hard and is probably not worth the effort ...
            content_height -= self.border_style.border_width

        if self.first_render_element:
            available_height = container_height - offset_y
            self.first_render_element = False
            rendering_complete = self.container.create_render_elements(
                container_top + offset_y, content_height, ctx, pdf_doc)

            needed_height = self.get_used_height()

            if rendering_complete and needed_height <= available_height:
                # rendering is complete and all elements of frame fit on current page
                self.rendering_complete = True
                self.render_bottom = offset_y + needed_height
                self.render_element_type = RenderElementType.complete
                render_element.add_elements(self.container, self.render_element_type, needed_height)
                return render_element, True
            else:
                if offset_y == 0:
                    # rendering of frame elements does not fit on current page but
                    # we are already at the top of the page -> start rendering and continue on next page
                    self.render_bottom = offset_y + available_height
                    self.render_element_type = RenderElementType.first
                    render_element.add_elements(self.container, self.render_element_type, available_height)
                    return render_element, False
                else:
                    # rendering of frame elements does not fit on current page -> start rendering on next page
                    self.next_page_rendering_complete = rendering_complete
                    self.prev_page_content_height = content_height
                    return None, False

        if self.render_element_type == RenderElementType.none:
            # render elements were already created on first call to get_next_render_element
            # but elements did not fit on first page

            if content_height == self.prev_page_content_height:
                # we don't have to create any render elements here because we can use
                # the previously created elements

                self.rendering_complete = self.next_page_rendering_complete
            else:
                # we cannot use previously created render elements because container height is different
                # on current page. this should be very unlikely but could happen when the frame should be
                # printed on the first page and header/footer are not shown on first page, i.e. the following
                # pages have a different content band size than the first page.

                self.container.prepare(ctx, pdf_doc=pdf_doc)
                self.rendering_complete = self.container.create_render_elements(
                    container_top, content_height, ctx, pdf_doc)
        else:
            self.rendering_complete = self.container.create_render_elements(
                container_top, content_height, ctx, pdf_doc)
        self.render_bottom = offset_y + self.get_used_height()

        if not self.rendering_complete:
            # use whole size of container if frame is not rendered completely
            self.render_bottom = offset_y + container_height

            if self.render_element_type == RenderElementType.none:
                self.render_element_type = RenderElementType.first
            else:
                self.render_element_type = RenderElementType.between
        else:
            if self.render_element_type == RenderElementType.none:
                self.render_element_type = RenderElementType.complete
            else:
                self.render_element_type = RenderElementType.last
        render_element.add_elements(self.container, self.render_element_type, self.get_used_height())
        return render_element, self.rendering_complete

    def render_spreadsheet(self, row, col, ctx, renderer):
        if self.spreadsheet_column:
            col = self.spreadsheet_column - 1
        row, col = self.container.render_spreadsheet(row, col, ctx, renderer)
        if self.spreadsheet_add_empty_row:
            row += 1
        return row, col


class SectionBandElement(object):
    def __init__(self, report, data, band_type, containers):
        from .containers import Container
        assert(isinstance(data, dict))
        self.id = get_int_value(data, 'id')
        self.width = report.document_properties.page_width -\
            report.document_properties.margin_left - report.document_properties.margin_right
        self.height = get_int_value(data, 'height')
        self.band_type = band_type
        if band_type == BandType.header:
            self.repeat_header = bool(data.get('repeatHeader'))
            self.always_print_on_same_page = True
        else:
            self.repeat_header = None
            self.always_print_on_same_page = bool(data.get('alwaysPrintOnSamePage'))
        self.shrink_to_content_height = bool(data.get('shrinkToContentHeight'))

        self.container = Container(
            container_id=str(data.get('linkedContainerId')), containers=containers, report=report)
        self.container.width = self.width
        self.container.height = self.height
        if band_type != BandType.content:
            self.container.allow_page_break = False
        self.rendering_complete = False
        self.prepare_container = True
        self.rendered_band_height = 0

    def prepare(self, ctx, pdf_doc, only_verify):
        pass

    def create_render_elements(self, offset_y, container_top, container_height, ctx, pdf_doc):
        available_height = container_height - offset_y
        if self.always_print_on_same_page and not self.shrink_to_content_height and\
                available_height < self.height:
            # not enough space for whole band
            self.rendering_complete = False
        else:
            if self.prepare_container:
                self.container.prepare(ctx, pdf_doc)
                self.rendered_band_height = 0
            else:
                self.rendered_band_height += self.container.used_band_height
                # clear render elements from previous page
                self.container.clear_rendered_elements()
            self.rendering_complete = self.container.create_render_elements(
                container_top + offset_y, available_height, ctx=ctx, pdf_doc=pdf_doc)

            if self.container.manual_page_break and self.always_print_on_same_page:
                # a manual page break is not allowed if the whole content should be printed on the same page
                raise ReportBroError(
                    Error('errorMsgSectionBandPageBreakNotAllowed', object_id=self.id, field='alwaysPrintOnSamePage'))

        if self.rendering_complete:
            remaining_min_height = self.height - self.rendered_band_height - self.container.skipped_height
            if not self.shrink_to_content_height and self.container.used_band_height < remaining_min_height:
                # rendering of band complete, make sure band is at least as large
                # as minimum height (even if it spans over more than 1 page)
                if remaining_min_height <= available_height:
                    self.prepare_container = True
                    self.container.used_band_height = remaining_min_height
                else:
                    # minimum height is larger than available space, continue on next page
                    self.rendering_complete = False
                    self.prepare_container = False
                    self.container.used_band_height = available_height
            else:
                self.prepare_container = True
        else:
            if self.always_print_on_same_page:
                # band must be printed on same page but available space is not enough,
                # try to render it on top of next page
                self.prepare_container = True
                if offset_y == 0:
                    field = 'size' if self.band_type == BandType.header else 'alwaysPrintOnSamePage'
                    raise ReportBroError(
                        Error('errorMsgSectionBandNotOnSamePage', object_id=self.id, field=field))
            else:
                self.prepare_container = False
                if self.container.manual_page_break:
                    # in case of manual page break the used band height
                    # is the position of the page break element
                    self.container.used_band_height = self.container.first_element_offset_y
                else:
                    # in case the available height was not sufficient the used band height
                    # is the total available height
                    self.container.used_band_height = available_height

    def get_used_band_height(self):
        return self.container.used_band_height

    def get_render_elements(self):
        return self.container.render_elements


class SectionElement(DocElement):
    def __init__(self, report, data, containers):
        DocElement.__init__(self, report, data)
        self.data_source = get_str_value(data, 'dataSource')
        self.print_if = get_str_value(data, 'printIf')
        self.spreadsheet_hide = False

        header = bool(data.get('header'))
        footer = bool(data.get('footer'))
        if header:
            self.header = SectionBandElement(report, data.get('headerData'), BandType.header, containers)
        else:
            self.header = None
        self.content = SectionBandElement(report, data.get('contentData'), BandType.content, containers)
        if footer:
            self.footer = SectionBandElement(report, data.get('footerData'), BandType.footer, containers)
        else:
            self.footer = None
        self.print_header = self.header is not None

        self.x = 0
        self.width = 0
        self.height = self.content.height
        if self.header:
            self.height += self.header.height
        if self.footer:
            self.height += self.footer.height
        self.bottom = self.y + self.height

        self.row_parameters = dict()
        self.rows = []
        self.row_count = 0
        self.row_index = -1

    def prepare(self, ctx, pdf_doc, only_verify):
        parameter_name = Context.strip_parameter_name(self.data_source)
        param_ref = ctx.get_parameter(parameter_name)
        if param_ref is None:
            raise ReportBroError(
                Error('errorMsgMissingDataSourceParameter', object_id=self.id, field='dataSource'))
        data_source_parameter = param_ref.parameter
        if data_source_parameter.type != ParameterType.array:
            raise ReportBroError(
                Error('errorMsgInvalidDataSourceParameter', object_id=self.id, field='dataSource'))
        for row_parameter in data_source_parameter.children:
            self.row_parameters[row_parameter.name] = row_parameter
        self.rows, parameter_exists = ctx.get_parameter_data(param_ref)
        if not parameter_exists:
            raise ReportBroError(
                Error('errorMsgMissingData', object_id=self.id, field='dataSource'))
        if not isinstance(self.rows, list):
            raise ReportBroError(
                Error('errorMsgInvalidDataSource', object_id=self.id, field='dataSource'))

        self.row_count = len(self.rows)
        self.row_index = 0

        if only_verify:
            if self.header:
                self.header.prepare(ctx, pdf_doc=None, only_verify=True)
            while self.row_index < self.row_count:
                # push data context of current row so values of current row can be accessed
                ctx.push_context(self.row_parameters, self.rows[self.row_index])
                self.content.prepare(ctx, pdf_doc=None, only_verify=True)
                ctx.pop_context()
                self.row_index += 1
            if self.footer:
                self.footer.prepare(ctx, pdf_doc=None, only_verify=True)

    def get_next_render_element(self, offset_y, container_top, container_height, ctx, pdf_doc):
        self.render_y = offset_y
        self.render_bottom = self.render_y
        render_element = SectionRenderElement(self.report, render_y=offset_y)

        if self.print_header:
            self.header.create_render_elements(offset_y, container_top, container_height, ctx, pdf_doc)
            render_element.add_section_band(self.header)
            if not self.header.rendering_complete:
                return render_element, False
            if not self.header.repeat_header:
                self.print_header = False

        while self.row_index < self.row_count:
            # push data context of current row so values of current row can be accessed
            ctx.push_context(self.row_parameters, self.rows[self.row_index])
            self.content.create_render_elements(
                offset_y + render_element.height, container_top, container_height, ctx, pdf_doc)
            ctx.pop_context()
            render_element.add_section_band(self.content)
            if not self.content.rendering_complete:
                return render_element, False
            self.row_index += 1

            page_break = self.content.container.manual_page_break
            self.content.container.reset()

            # in case of a manual page break inside content band (unless already in last row)
            # we stop rendering and start on new page
            if page_break and self.row_index < self.row_count:
                return render_element, False

        if self.footer:
            self.footer.create_render_elements(
                offset_y + render_element.height, container_top, container_height, ctx, pdf_doc)
            render_element.add_section_band(self.footer)
            if not self.footer.rendering_complete:
                return render_element, False

        # all bands finished
        self.rendering_complete = True
        self.render_bottom += render_element.height
        return render_element, True

    def render_spreadsheet(self, row, col, ctx, renderer):
        if self.header:
            self.header.container.prepare(ctx, pdf_doc=None)
            row, _ = self.header.container.render_spreadsheet(row, col, ctx, renderer)

        while self.row_index < self.row_count:
            # push data context of current row so values of current row can be accessed
            ctx.push_context(self.row_parameters, self.rows[self.row_index])
            self.content.container.prepare(ctx, pdf_doc=None)
            row, _ = self.content.container.render_spreadsheet(row, col, ctx, renderer)
            self.row_index += 1

        if self.footer:
            self.footer.container.prepare(ctx, pdf_doc=None)
            row, _ = self.footer.container.render_spreadsheet(row, col, ctx, renderer)
        return row, col



from __future__ import unicode_literals
from __future__ import division
from babel.numbers import format_decimal
from babel.dates import format_datetime
from collections import namedtuple
from simpleeval import simple_eval, NameNotDefined, FunctionNotDefined
from simpleeval import DEFAULT_NAMES as EVAL_DEFAULT_NAMES
from simpleeval import DEFAULT_FUNCTIONS as EVAL_DEFAULT_FUNCTIONS
from typing import List
import datetime
import decimal

from .enums import *
from .errors import Error, ReportBroError


# parameter instance, the data map referenced by the parameter and the data map containing
# the context_id (this is usually the data map but can be different for collection
# parameters)
ParameterRef = namedtuple('ParameterRef', ['parameter', 'data', 'data_context'])
ContextEntry = namedtuple('ContextEntry', ['parameters', 'data', 'prev_entry'])
CONTEXT_ENTRY_PARAMETERS = 0
CONTEXT_ENTRY_DATA = 1
CONTEXT_ENTRY_PREV_ENTRY = 2


class Context:
    def __init__(self, report, parameters, data):
        self.report = report
        self.pattern_locale = report.document_properties.pattern_locale
        self.pattern_currency_symbol = report.document_properties.pattern_currency_symbol
        data.update(EVAL_DEFAULT_NAMES)
        # each new context (push_context) gets a new unique id
        self.id = 1
        data['__context_id'] = self.id
        self.eval_functions = EVAL_DEFAULT_FUNCTIONS.copy()
        self.eval_functions.update(
            len=len,
            decimal=decimal.Decimal,
            datetime=datetime
        )
        self.root_data = data
        self.root_data['page_number'] = 0
        self.root_data['page_count'] = 0
        self.context_stack: List[ContextEntry] = [ContextEntry(parameters=parameters, data=data, prev_entry=None)]
        # a range count is increased inside a table group band (e.g. to show header or sums for grouped rows),
        # if a range is set we have to evaluate parameter functions (e.g. sum/avg) because the range could be affected
        self.range_count = 0

    def get_parameter(self, name):
        """
        Return parameter reference for given parameter name.

        :param name: name of the parameter to find, the parameter can be present in the current
        context or any of its parents.
        :return: parameter reference which contains a parameter instance and
        its data map referenced by the parameter. None if no parameter was found.
        """
        if name.find('.') != -1:
            # this parameter is part of a collection, so we first get the reference to the
            # collection parameter and then return the parameter inside the collection
            name_parts = name.split('.')
            collection_name = name_parts[0]
            field_name = name_parts[1]
            param_ref = self._get_parameter(collection_name)
            if param_ref is not None and param_ref.parameter.type == ParameterType.map and\
                    field_name in param_ref.parameter.fields and collection_name in param_ref.data:
                return ParameterRef(
                    parameter=param_ref.parameter.fields[field_name],
                    data=param_ref.data[collection_name], data_context=param_ref.data)
            return None
        else:
            return self._get_parameter(name)

    def _get_parameter(self, name, context_entry=None):
        if context_entry is None:
            context_entry = self.context_stack[-1]
        parameters = context_entry[CONTEXT_ENTRY_PARAMETERS]

        if name in parameters:
            data = context_entry[CONTEXT_ENTRY_DATA]
            return ParameterRef(parameter=parameters[name], data=data, data_context=data)
        else:
            parent_context_entry = context_entry[CONTEXT_ENTRY_PREV_ENTRY]
            if parent_context_entry:
                return self._get_parameter(name, context_entry=parent_context_entry)
        return None

    def get_parameter_data(self, param_ref):
        """
        Return data for given parameter reference.

        :param param_ref: a parameter reference which contains a parameter instance and
        its data map referenced by the parameter.
        :return: tuple of current data value of parameter, bool if parameter data exists
        """
        parameter = param_ref.parameter
        if self.range_count and parameter.is_range_function():
            return self.evaluate_parameter_func(parameter)
        elif parameter.name in param_ref.data:
            return param_ref.data[parameter.name], True
        return None, False

    @staticmethod
    def get_parameter_context_id(param_ref):
        """
        Return context_id for given parameter reference.

        This can be useful to find out if a parameter value has changed,
        e.g. parameter 'amount' in a list of invoice items has a different context_id
        in each list row (invoice item).

        :param param_ref: a parameter reference which contains a parameter instance and
        its data map referenced by the parameter.
        :return: unique context id or None if there is no context available.
        """
        if '__context_id' in param_ref.data_context:
            return param_ref.data_context['__context_id']
        return None

    def push_context(self, parameters, data):
        self.id += 1
        data['__context_id'] = self.id
        current_context_entry = self.context_stack[-1]
        self.context_stack.append(ContextEntry(parameters=parameters, data=data, prev_entry=current_context_entry))

    def pop_context(self):
        if len(self.context_stack) <= 1:
            raise RuntimeError('Context.pop_context failed')
        self.context_stack = self.context_stack[:-1]

    def fill_parameters(self, expr, object_id, field, pattern=None):
        if expr.find('${') == -1:
            return expr
        rv = ''
        prev_c = None
        parameter_index = -1
        for i, c in enumerate(expr):
            if parameter_index == -1:
                if prev_c == '$' and c == '{':
                    parameter_index = i + 1
                    rv = rv[:-1]
                else:
                    rv += c
            else:
                if c == '}':
                    parameter_name = expr[parameter_index:i]
                    param_ref = self.get_parameter(parameter_name)
                    if param_ref is None:
                        raise ReportBroError(
                            Error('errorMsgInvalidExpressionNameNotDefined',
                                  object_id=object_id, field=field, info=parameter_name))
                    value, value_exists = self.get_parameter_data(param_ref)

                    if not value_exists:
                        raise ReportBroError(
                            Error('errorMsgMissingParameterData',
                                  object_id=object_id, field=field, info=parameter_name))

                    if value is not None:
                        rv += self.get_formatted_value(value, param_ref.parameter, object_id, pattern=pattern)
                    parameter_index = -1
            prev_c = c
        return rv

    def evaluate_expression(self, expr, object_id, field):
        if expr:
            try:
                data = dict(EVAL_DEFAULT_NAMES)
                expr = self.replace_parameters(expr, data=data)
                return simple_eval(expr, names=data, functions=self.eval_functions)
            except NameNotDefined as ex:
                raise ReportBroError(
                    Error('errorMsgInvalidExpressionNameNotDefined',
                          object_id=object_id, field=field, info=ex.name, context=expr))
            except FunctionNotDefined as ex:
                # avoid possible unresolved attribute reference warning by using getattr
                func_name = getattr(ex, 'func_name')
                raise ReportBroError(
                    Error('errorMsgInvalidExpressionFuncNotDefined',
                          object_id=object_id, field=field, info=func_name, context=expr))
            except SyntaxError as ex:
                raise ReportBroError(
                    Error('errorMsgInvalidExpression', object_id=object_id, field=field, info=ex.msg, context=expr))
            except Exception as ex:
                info = ex.message if hasattr(ex, 'message') else str(ex)
                raise ReportBroError(
                    Error('errorMsgInvalidExpression', object_id=object_id, field=field, info=info, context=expr))
        return True

    def evaluate_parameter_func(self, parameter):
        expr = Context.strip_parameter_name(parameter.expression)
        pos = expr.find('.')
        if pos == -1:
            raise ReportBroError(
                Error('errorMsgInvalidAvgSumExpression',
                      object_id=parameter.id, field='expression', context=parameter.name))
        else:
            parameter_name = expr[:pos]
            parameter_field = expr[pos+1:]
            param_ref = self.get_parameter(parameter_name)
            if param_ref is None or param_ref.parameter.type != ParameterType.array:
                raise ReportBroError(
                    Error('errorMsgInvalidAvgSumExpression',
                          object_id=parameter.id, field='expression', context=parameter.name))
            else:
                total = decimal.Decimal(0)
                items, data_exists = self.get_parameter_data(param_ref)
                if not data_exists or not isinstance(items, list):
                    raise ReportBroError(
                        Error('errorMsgInvalidAvgSumExpression',
                              object_id=parameter.id, field='expression', context=parameter.name))

                start, end = param_ref.parameter.get_range()
                if start is not None:
                    items = items[start:(end if end != -1 else len(items))]

                for item in items:
                    item_value = item.get(parameter_field)
                    if not isinstance(item_value, decimal.Decimal):
                        raise ReportBroError(
                            Error('errorMsgInvalidAvgSumExpression',
                                  object_id=parameter.id, field='expression', context=parameter.name))
                    total += item_value

                value = None
                if parameter.type == ParameterType.average:
                    value = total / len(items)
                elif parameter.type == ParameterType.sum:
                    value = total
                return value, True

    @staticmethod
    def strip_parameter_name(expr):
        if expr:
            return expr.strip().lstrip('${').rstrip('}')
        return expr

    @staticmethod
    def is_parameter_name(expr):
        return expr and expr.lstrip().startswith('${') and expr.rstrip().endswith('}')

    def get_formatted_value(self, value, parameter, object_id, pattern=None, is_array_item=False):
        rv = ''
        if is_array_item and parameter.type == ParameterType.simple_array:
            value_type = parameter.array_item_type
        else:
            value_type = parameter.type
        if value_type == ParameterType.string:
            rv = value
        elif value_type in (ParameterType.number, ParameterType.average, ParameterType.sum):
            if pattern:
                used_pattern = pattern
                pattern_has_currency = (pattern.find('$') != -1)
            else:
                used_pattern = parameter.pattern
                pattern_has_currency = parameter.pattern_has_currency
            if used_pattern:
                try:
                    value = format_decimal(value, used_pattern, locale=self.pattern_locale)
                    if pattern_has_currency:
                        value = value.replace('$', self.pattern_currency_symbol)
                    rv = value
                except ValueError:
                    error_object_id = object_id if pattern else parameter.id
                    raise ReportBroError(
                        Error('errorMsgInvalidPattern', object_id=error_object_id, field='pattern', context=value))
            else:
                rv = str(value)
        elif value_type == ParameterType.date:
            used_pattern = pattern if pattern else parameter.pattern
            if used_pattern:
                try:
                    rv = format_datetime(value, used_pattern, locale=self.pattern_locale)
                except ValueError:
                    error_object_id = object_id if pattern else parameter.id
                    raise ReportBroError(
                        Error('errorMsgInvalidPattern',
                              object_id=error_object_id, field='pattern', context=value))
            else:
                rv = str(value)
        return rv

    def replace_parameters(self, expr, data):
        pos = expr.find('${')
        if pos == -1:
            return expr
        rv = ''
        pos2 = 0
        while pos != -1:
            if pos != 0:
                rv += expr[pos2:pos]
            pos2 = expr.find('}', pos)
            if pos2 != -1:
                parameter_name = expr[pos+2:pos2]

                param_ref = self.get_parameter(parameter_name)
                parameter_name = parameter_name.replace('.', '_')
                if param_ref:
                    value, _ = self.get_parameter_data(param_ref)
                else:
                    value = None

                data[parameter_name] = value
                rv += parameter_name
                pos2 += 1
                pos = expr.find('${', pos2)
            else:
                pos2 = pos
                pos = -1
        rv += expr[pos2:]
        return rv

    def inc_range_count(self):
        self.range_count += 1

    def dec_range_count(self):
        self.range_count -= 1

    def inc_page_number(self):
        self.root_data['page_number'] += 1

    def get_page_number(self):
        return self.root_data['page_number']

    def set_page_count(self, page_count):
        self.root_data['page_count'] = page_count



from __future__ import unicode_literals
from __future__ import division
from .enums import *
from .errors import Error
from .utils import get_float_value, get_int_value


class Color:
    def __init__(self, color):
        self.color_code = ''
        if color:
            assert len(color) == 7 and color[0] == '#'
            self.r = int(color[1:3], 16)
            self.g = int(color[3:5], 16)
            self.b = int(color[5:7], 16)
            self.transparent = False
            self.color_code = color
        else:
            self.r = 0
            self.g = 0
            self.b = 0
            self.transparent = True
            self.color_code = ''

    def __eq__(self, other):
        if isinstance(other, Color):
            return self.color_code == other.color_code
        return False

    def is_black(self):
        return self.r == 0 and self.g == 0 and self.b == 0 and not self.transparent


class Parameter:
    def __init__(self, report, data):
        self.report = report
        self.id = int(data.get('id'))
        self.name = data.get('name', '<unnamed>')
        self.type = ParameterType[data.get('type')]
        if self.type == ParameterType.simple_array:
            self.array_item_type = ParameterType[data.get('arrayItemType')]
        else:
            self.array_item_type = ParameterType.none
        self.eval = bool(data.get('eval'))
        self.nullable = bool(data.get('nullable'))
        self.expression = data.get('expression', '')
        self.pattern = data.get('pattern', '')
        self.pattern_has_currency = (self.pattern.find('$') != -1)
        self.is_internal = self.name in ('page_count', 'page_number', 'row_number')
        self.range_stack = []
        self.children = []
        self.fields = dict()
        if self.type == ParameterType.array or self.type == ParameterType.map:
            for item in data.get('children'):
                parameter = Parameter(self.report, item)
                if parameter.name in self.fields:
                    self.report.errors.append(
                        Error('errorMsgDuplicateParameterField', object_id=parameter.id, field='name'))
                else:
                    self.children.append(parameter)
                    self.fields[parameter.name] = parameter

    def is_evaluated(self):
        """Return True if parameter data must be evaluated initially."""
        return self.eval or self.is_range_function()

    def is_range_function(self):
        """Return True if parameter is a function with range input."""
        return self.type in (ParameterType.average, ParameterType.sum)

    def set_range(self, row_start, row_end):
        """
        Set row range which is used for parameter functions (e.g. sum/avg), if a range is set then
        only these rows will be used for the function, otherwise all rows are used.

        :param row_start: first row of group
        :param row_end: index after last row of group
        """
        self.range_stack.append((row_start, row_end))

    def clear_range(self):
        """Clear previously set row range."""
        self.range_stack.pop()

    def get_range(self):
        if self.range_stack:
            return self.range_stack[-1]
        return None, None

    def has_range(self):
        return bool(self.range_stack)


class BorderStyle:
    def __init__(self, data, key_prefix=''):
        self.border_color = Color(data.get(key_prefix + 'borderColor'))
        self.border_width = get_float_value(data, key_prefix + 'borderWidth')
        self.border_all = bool(data.get(key_prefix + 'borderAll'))
        self.border_left = self.border_all or bool(data.get(key_prefix + 'borderLeft'))
        self.border_top = self.border_all or bool(data.get(key_prefix + 'borderTop'))
        self.border_right = self.border_all or bool(data.get(key_prefix + 'borderRight'))
        self.border_bottom = self.border_all or bool(data.get(key_prefix + 'borderBottom'))


class TextStyle(BorderStyle):
    def __init__(self, data, key_prefix=''):
        BorderStyle.__init__(self, data, key_prefix)
        self.key_prefix = key_prefix
        self.id = str(get_int_value(data, 'id'))
        self.bold = bool(data.get(key_prefix + 'bold'))
        self.italic = bool(data.get(key_prefix + 'italic'))
        self.underline = bool(data.get(key_prefix + 'underline'))
        self.strikethrough = bool(data.get(key_prefix + 'strikethrough'))
        self.horizontal_alignment = HorizontalAlignment[data.get(key_prefix + 'horizontalAlignment')]
        self.vertical_alignment = VerticalAlignment[data.get(key_prefix + 'verticalAlignment')]
        self.text_color = Color(data.get(key_prefix + 'textColor'))
        self.background_color = Color(data.get(key_prefix + 'backgroundColor'))
        self.font = data.get(key_prefix + 'font')
        self.font_size = get_int_value(data, key_prefix + 'fontSize')
        self.line_spacing = get_float_value(data, key_prefix + 'lineSpacing')
        self.padding_left = get_int_value(data, key_prefix + 'paddingLeft')
        self.padding_top = get_int_value(data, key_prefix + 'paddingTop')
        self.padding_right = get_int_value(data, key_prefix + 'paddingRight')
        self.padding_bottom = get_int_value(data, key_prefix + 'paddingBottom')
        self.font_style = ''
        if self.bold:
            self.font_style += 'B'
        if self.italic:
            self.font_style += 'I'
        self.text_align = ''
        if self.horizontal_alignment == HorizontalAlignment.left:
            self.text_align = 'L'
        elif self.horizontal_alignment == HorizontalAlignment.center:
            self.text_align = 'C'
        elif self.horizontal_alignment == HorizontalAlignment.right:
            self.text_align = 'R'
        elif self.horizontal_alignment == HorizontalAlignment.justify:
            self.text_align = 'J'
        self.add_border_padding()

    def set_bold(self, bold):
        self.bold = bold
        self.font_style = self.get_font_style(ignore_underline=True)

    def set_italic(self, italic):
        self.italic = italic
        self.font_style = self.get_font_style(ignore_underline=True)

    def get_font_style(self, ignore_underline=False):
        font_style = ''
        if self.bold:
            font_style += 'B'
        if self.italic:
            font_style += 'I'
        if self.underline and not ignore_underline:
            font_style += 'U'
        return font_style

    def add_border_padding(self):
        if self.border_left:
            self.padding_left += self.border_width
        if self.border_top:
            self.padding_top += self.border_width
        if self.border_right:
            self.padding_right += self.border_width
        if self.border_bottom:
            self.padding_bottom += self.border_width



from .enums import *
from .utils import get_int_value


class DocElementBase(object):
    """Base class for all elements defined in the report template."""

    def __init__(self, report, data):
        self.report = report
        self.id = None
        self.y = get_int_value(data, 'y')
        self.render_y = 0
        self.render_bottom = 0
        self.bottom = self.y
        self.height = 0
        self.print_if = None
        self.remove_empty_element = False
        self.spreadsheet_hide = True
        self.spreadsheet_column = None
        self.spreadsheet_add_empty_row = False
        self.first_render_element = True
        self.rendering_complete = False
        self.predecessors = []
        self.successors = []
        self.sort_order = 1  # sort order for elements with same 'y'-value

    def is_predecessor(self, elem):
        """Returns true if the given element is a direct predecessor of this element.

        An element is a direct predecessor if it ends before or on the same position as
        this elements starts. Further it has to end after start of a possible
        existing predecessor, if this is not the case then the given element is already
        a predecessor of the existing predecessor.

        The current element can only be printed after all predecessors are finished.
        """
        return self.y >= elem.bottom and (len(self.predecessors) == 0 or elem.bottom > self.predecessors[0].y)

    def add_predecessor(self, predecessor):
        self.predecessors.append(predecessor)
        predecessor.successors.append(self)

    def has_uncompleted_predecessor(self, completed_elements):
        """returns True in case there is at least one predecessor which is not completely rendered yet."""
        for predecessor in self.predecessors:
            if predecessor.id not in completed_elements or not predecessor.rendering_complete:
                return True
        return False

    def get_offset_y(self):
        """Returns offset y-coord for rendering of this element.

        The value is calculated as lowest bottom (i.e. highest value) of predecessors plus
        minimum space to any predecessor.
        """
        max_bottom = 0
        min_predecessor_dist = None
        for predecessor in self.predecessors:
            if predecessor.render_bottom > max_bottom:
                max_bottom = predecessor.render_bottom
            predecessor_dist = (self.y - predecessor.bottom)
            if min_predecessor_dist is None or predecessor_dist < min_predecessor_dist:
                min_predecessor_dist = predecessor_dist
        return max_bottom + (min_predecessor_dist if min_predecessor_dist else 0)

    def clear_predecessor(self, elem):
        if elem in self.predecessors:
            self.predecessors.remove(elem)

    def prepare(self, ctx, pdf_doc, only_verify):
        pass

    def is_printed(self, ctx):
        if self.print_if:
            return ctx.evaluate_expression(self.print_if, self.id, field='printIf')
        return True

    def finish_empty_element(self, offset_y):
        if self.remove_empty_element:
            self.render_bottom = offset_y
        else:
            self.render_bottom = offset_y + self.height
        self.rendering_complete = True

    def get_next_render_element(self, offset_y, container_top, container_height, ctx, pdf_doc):
        self.rendering_complete = True
        return None, True

    def render_pdf(self, container_offset_x, container_offset_y, pdf_doc):
        pass

    def render_spreadsheet(self, row, col, ctx, renderer):
        return row, col

    def cleanup(self):
        pass


class DocElement(DocElementBase):
    def __init__(self, report, data):
        DocElementBase.__init__(self, report, data)
        self.id = get_int_value(data, 'id')
        self.x = get_int_value(data, 'x')
        self.width = get_int_value(data, 'width')
        self.height = get_int_value(data, 'height')
        self.bottom = self.y + self.height

    def get_next_render_element(self, offset_y, container_top, container_height, ctx, pdf_doc):
        if offset_y + self.height <= container_height:
            self.render_y = offset_y
            self.render_bottom = offset_y + self.height
            self.rendering_complete = True
            return self, True
        return None, False

    @staticmethod
    def draw_border(x, y, width, height, render_element_type, border_style, pdf_doc):
        pdf_doc.set_draw_color(
            border_style.border_color.r, border_style.border_color.g, border_style.border_color.b)
        pdf_doc.set_line_width(border_style.border_width)
        border_offset = border_style.border_width / 2
        border_x = x + border_offset
        border_y = y + border_offset
        border_width = width - border_style.border_width
        border_height = height - border_style.border_width
        if border_style.border_all and render_element_type == RenderElementType.complete:
            pdf_doc.rect(border_x, border_y, border_width, border_height, style='D')
        else:
            if border_style.border_left:
                pdf_doc.line(border_x, border_y, border_x, border_y + border_height)
            if border_style.border_top and render_element_type in (
                        RenderElementType.complete, RenderElementType.first):
                pdf_doc.line(border_x, border_y, border_x + border_width, border_y)
            if border_style.border_right:
                pdf_doc.line(
                    border_x + border_width, border_y,
                    border_x + border_width, border_y + border_height)
            if border_style.border_bottom and render_element_type in (
                        RenderElementType.complete, RenderElementType.last):
                pdf_doc.line(
                    border_x, border_y + border_height,
                    border_x + border_width, border_y + border_height)



from typing import List
from .elements import DocElementBase, PageBreakElement
from .enums import BandType


class Container(object):
    def __init__(self, container_id, containers, report):
        self.id = container_id
        self.report = report
        self.doc_elements = []  # type: List[DocElementBase]
        self.width = 0
        self.height = 0
        if containers is not None:
            containers[self.id] = self

        self.allow_page_break = True
        self.container_offset_y = 0
        self.sorted_elements = None  # type: List[DocElementBase]
        self.render_elements = None  # type: List[DocElementBase]
        self.render_elements_created = False
        self.manual_page_break = False
        self.first_element_offset_y = 0
        # in case not all elements could be rendered on the same page, the next element will start
        # at the top of the next page independent of the actual distance to its predecessor. the
        # skipped height is the difference to the predecessor which gets "lost" because of the new page
        self.skipped_height = 0
        self.used_band_height = 0

    def add(self, doc_element):
        self.doc_elements.append(doc_element)

    def is_visible(self):
        return True

    def prepare(self, ctx, pdf_doc=None, only_verify=False):
        self.sorted_elements = []
        for elem in self.doc_elements:
            if pdf_doc or not elem.spreadsheet_hide or only_verify:
                elem.prepare(ctx, pdf_doc=pdf_doc, only_verify=only_verify)
                if not self.allow_page_break:
                    # make sure element can be rendered multiple times (for header/footer)
                    elem.first_render_element = True
                    elem.rendering_complete = False
                # if page break is not printed we have to skip it during prepare because
                # offset calculations between elements are affected
                if not isinstance(elem, PageBreakElement) or elem.is_printed(ctx):
                    self.sorted_elements.append(elem)

        if pdf_doc:
            self.sorted_elements = sorted(self.sorted_elements, key=lambda item: (item.y, item.sort_order))
            # predecessors are only needed for rendering pdf document
            for i, elem in enumerate(self.sorted_elements):
                for j in range(i-1, -1, -1):
                    elem2 = self.sorted_elements[j]
                    if isinstance(elem2, PageBreakElement):
                        # new page so all elements before are not direct predecessors
                        break
                    if elem.is_predecessor(elem2):
                        elem.add_predecessor(elem2)
            self.render_elements = []
            self.used_band_height = 0
            self.first_element_offset_y = 0
        else:
            self.sorted_elements = sorted(self.sorted_elements, key=lambda item: (item.y, item.x))

    def clear_rendered_elements(self):
        self.render_elements = []
        self.used_band_height = 0

    def get_render_elements_bottom(self):
        bottom = 0
        for render_element in self.render_elements:
            if render_element.render_bottom > bottom:
                bottom = render_element.render_bottom
        return bottom

    def create_render_elements(self, container_top, container_height, ctx, pdf_doc):
        i = 0
        new_page = False
        processed_elements = []
        completed_elements = dict()

        self.render_elements_created = False
        self.manual_page_break = False
        next_offset_y = None
        while not new_page and i < len(self.sorted_elements):
            elem = self.sorted_elements[i]
            if elem.has_uncompleted_predecessor(completed_elements):
                # a predecessor is not completed yet -> start new page
                new_page = True
            else:
                elem_deleted = False
                if isinstance(elem, PageBreakElement):
                    if self.allow_page_break:
                        del self.sorted_elements[i]
                        elem_deleted = True
                        new_page = True
                        self.manual_page_break = True
                        next_offset_y = elem.y
                    else:
                        self.sorted_elements = []
                        return True
                else:
                    complete = False
                    if elem.predecessors:
                        # element is on same page as predecessor element(s) so offset is relative to predecessors
                        offset_y = elem.get_offset_y()
                    else:
                        if elem.first_render_element:
                            offset_y = elem.y - self.first_element_offset_y
                            if offset_y < 0:
                                offset_y = 0
                        else:
                            offset_y = 0

                    if elem.is_printed(ctx):
                        if offset_y >= container_height:
                            # add to "lost" height of container because next element is printed at top
                            # of next page instead of actual offset
                            self.skipped_height += offset_y - container_height
                            new_page = True
                        if not new_page:
                            render_elem, complete = elem.get_next_render_element(
                                offset_y, container_top=container_top,
                                container_height=container_height, ctx=ctx, pdf_doc=pdf_doc)
                            if complete:
                                processed_elements.append(elem)
                            if render_elem:
                                self.render_elements.append(render_elem)
                                self.render_elements_created = True
                                if render_elem.render_bottom > self.used_band_height:
                                    self.used_band_height = render_elem.render_bottom
                    else:
                        processed_elements.append(elem)
                        elem.finish_empty_element(offset_y)
                        complete = True

                    if not complete and next_offset_y is None:
                        # in case we continue rendering on the next page the first element which is not complete
                        # will define the offset-y for the next page
                        next_offset_y = elem.y

                    if complete:
                        completed_elements[elem.id] = True
                        del self.sorted_elements[i]
                        elem_deleted = True
                if not elem_deleted:
                    i += 1

        self.first_element_offset_y = next_offset_y if next_offset_y else 0

        if len(self.sorted_elements) > 0:
            if self.allow_page_break:
                self.render_elements.append(PageBreakElement(self.report, dict(y=-1)))
            for processed_element in processed_elements:
                # remove dependency to predecessors because successor element is either already added
                # to render_elements or on new page
                for successor in processed_element.successors:
                    successor.clear_predecessor(processed_element)
        return len(self.sorted_elements) == 0

    def render_pdf(self, container_offset_x, container_offset_y, pdf_doc, cleanup=False):
        counter = 0
        for render_elem in self.render_elements:
            counter += 1
            if isinstance(render_elem, PageBreakElement):
                break
            render_elem.render_pdf(container_offset_x, container_offset_y, pdf_doc)
            if cleanup:
                render_elem.cleanup()
        self.render_elements = self.render_elements[counter:]

    def render_spreadsheet(self, row, col, ctx, renderer):
        max_col = col
        i = 0
        count = len(self.sorted_elements)
        while i < count:
            elem = self.sorted_elements[i]
            if elem.is_printed(ctx):
                j = i + 1
                # render elements with same y-coordinate in same spreadsheet row
                row_elements = [elem]
                while j < count:
                    elem2 = self.sorted_elements[j]
                    if elem2.y == elem.y:
                        if elem2.is_printed(ctx):
                            row_elements.append(elem2)
                    else:
                        break
                    j += 1
                i = j
                current_row = row
                current_col = col
                for row_element in row_elements:
                    tmp_row, current_col = row_element.render_spreadsheet(
                        current_row, current_col, ctx, renderer)
                    row = max(row, tmp_row)
                    if current_col > max_col:
                        max_col = current_col
            else:
                i += 1
        return row, max_col

    def is_finished(self):
        return len(self.render_elements) == 0

    def reset(self):
        """Reset container when used multiple times.

        Must be called when the same container is used for rendering, e.g. for
        different rows in a section content band or a repeated header/footer.
        """
        self.manual_page_break = False
        self.first_element_offset_y = 0
        self.skipped_height = 0
        self.used_band_height = 0
        for elem in self.doc_elements:
            elem.first_render_element = True
            elem.rendering_complete = False

    def cleanup(self):
        for elem in self.doc_elements:
            elem.cleanup()


class Frame(Container):
    def __init__(self, width, height, container_id, containers, report):
        Container.__init__(self, container_id, containers, report)
        self.width = width
        self.height = height
        self.allow_page_break = False


class ReportBand(Container):
    def __init__(self, band, container_id, containers, report):
        Container.__init__(self, container_id, containers, report)
        self.band = band
        self.width = report.document_properties.page_width -\
            report.document_properties.margin_left - report.document_properties.margin_right
        if band == BandType.content:
            self.height = report.document_properties.content_height
        elif band == BandType.header:
            self.allow_page_break = False
            self.height = report.document_properties.header_size
        elif band == BandType.footer:
            self.allow_page_break = False
            self.height = report.document_properties.footer_size

    def is_visible(self):
        if self.band == BandType.header:
            return self.report.document_properties.header
        elif self.band == BandType.footer:
            return self.report.document_properties.footer
        return True



from __future__ import division
from .docelement import DocElementBase, DocElement
from .enums import *
from .errors import Error, ReportBroError
from .utils import get_image_display_size
import os


class ImageRenderElement(DocElementBase):
    def __init__(self, report, render_y, image):
        DocElementBase.__init__(self, report, dict(y=0))
        self.report = report
        self.x = image.x
        self.width = image.width
        self.height = image.height
        self.render_y = render_y
        self.render_bottom = render_y
        self.background_color = image.background_color
        self.horizontal_alignment = image.horizontal_alignment
        self.vertical_alignment = image.vertical_alignment
        self.prepared_link = image.prepared_link
        self.source = image.source
        self.image_filename = image.image_filename
        self.image_key = image.image_key

    def render_pdf(self, container_offset_x, container_offset_y, pdf_doc):
        x = self.x + container_offset_x
        y = self.render_y + container_offset_y
        if not self.background_color.transparent:
            pdf_doc.set_fill_color(self.background_color.r, self.background_color.g, self.background_color.b)
            pdf_doc.rect(x, y, self.width, self.height, 'F')
        if self.image_key:
            image = self.report.get_image(self.image_key)
            if image and image.image_fp:
                halign = {
                    HorizontalAlignment.left: 'L', HorizontalAlignment.center: 'C',
                    HorizontalAlignment.right: 'R'}.get(self.horizontal_alignment)
                valign = {
                    VerticalAlignment.top: 'T', VerticalAlignment.middle: 'C',
                    VerticalAlignment.bottom: 'B'}.get(self.vertical_alignment)
                try:
                    image_info = pdf_doc.image(
                        self.image_key, x, y, self.width, self.height, type=image.image_type,
                        image_fp=image.image_fp, halign=halign, valign=valign)
                except Exception as ex:
                    raise ReportBroError(
                        Error('errorMsgLoadingImageFailed', object_id=self.id,
                              field='source' if self.source else 'image', info=str(ex)))

                if self.prepared_link:
                    # horizontal and vertical alignment of image within given width and height
                    # by keeping original image aspect ratio
                    offset_x = offset_y = 0
                    image_display_width, image_display_height = get_image_display_size(
                        self.width, self.height, image_info['w'], image_info['h'])
                    if self.horizontal_alignment == HorizontalAlignment.center:
                        offset_x = (self.width - image_display_width) / 2
                    elif self.horizontal_alignment == HorizontalAlignment.right:
                        offset_x = self.width - image_display_width
                    if self.vertical_alignment == VerticalAlignment.middle:
                        offset_y = (self.height - image_display_height) / 2
                    elif self.vertical_alignment == VerticalAlignment.bottom:
                        offset_y = self.height - image_display_height

                    pdf_doc.link(x + offset_x, y + offset_y,
                                 image_display_width, image_display_height, self.prepared_link)


class BarCodeRenderElement(DocElementBase):
    def __init__(self, report, render_y, barcode):
        DocElementBase.__init__(self, report, dict(y=0))
        self.report = report
        self.x = barcode.x
        self.width = barcode.width
        self.height = barcode.height
        self.render_y = render_y
        self.render_bottom = render_y
        self.content = barcode.prepared_content
        self.display_value = barcode.display_value
        self.image_key = barcode.image_key
        self.image_height = barcode.image_height

    def render_pdf(self, container_offset_x, container_offset_y, pdf_doc):
        x = self.x + container_offset_x
        y = self.render_y + container_offset_y
        if self.image_key:
            pdf_doc.image(self.image_key, x, y, self.width, self.image_height)
            if self.display_value:
                pdf_doc.set_font('courier', 'B', 18)
                pdf_doc.set_text_color(0, 0, 0)
                content_width = pdf_doc.get_string_width(self.content)
                offset_x = (self.width - content_width) / 2
                pdf_doc.text(x + offset_x, y + self.image_height + 20, self.content)

    def cleanup(self):
        if self.image_key:
            os.unlink(self.image_key)
            self.image_key = None


class TableRenderElement(DocElementBase):
    def __init__(self, report, table, render_y):
        DocElementBase.__init__(self, report, dict(y=0))
        self.report = report
        self.table = table
        self.x = table.x
        self.width = table.width
        self.render_y = render_y
        self.render_bottom = render_y
        self.height = 0
        self.bands = []
        self.complete = False

    def is_empty(self):
        return len(self.bands) == 0

    def add_band(self, band, row_index=-1):
        if band.rendering_complete or not band.always_print_on_same_page:
            band_height = band.get_used_band_height()
            background_color = band.background_color
            if band.band_type == BandType.content and not band.alternate_background_color.transparent and\
                    row_index % 2 == 1:
                background_color = band.alternate_background_color

            self.bands.append(dict(
                height=band_height, background_color=background_color,
                elements=list(band.get_render_elements()), cells=band.printed_cells))
            self.height += band_height
            self.render_bottom += band_height

    def render_pdf(self, container_offset_x, container_offset_y, pdf_doc):
        x = self.x + container_offset_x
        x1 = x
        x2 = x1 + self.width
        y = self.render_y + container_offset_y
        row_y = y
        for band in self.bands:
            background_color = band['background_color']
            if not background_color.transparent:
                pdf_doc.set_fill_color(
                    background_color.r, background_color.g, background_color.b)
                pdf_doc.rect(x, row_y, self.width, band['height'], style='F')

            for element in band['elements']:
                element.render_pdf(container_offset_x=x, container_offset_y=row_y, pdf_doc=pdf_doc)
            row_y += band['height']

        if not self.is_empty() and self.table.border != Border.none:
            pdf_doc.set_draw_color(
                self.table.border_color.r, self.table.border_color.g, self.table.border_color.b)
            pdf_doc.set_line_width(self.table.border_width)
            half_border_width = self.table.border_width / 2
            x1 += half_border_width
            x2 -= half_border_width
            y1 = y
            y2 = row_y
            if self.table.border in (Border.grid, Border.frame_row, Border.frame):
                # draw left and right table borders
                pdf_doc.line(x1, y1, x1, y2)
                pdf_doc.line(x2, y1, x2, y2)
            y = y1
            pdf_doc.line(x1, y1, x2, y1)
            if self.table.border != Border.frame:
                # draw lines between table rows
                for band in self.bands[:-1]:
                    y += band['height']
                    pdf_doc.line(x1, y, x2, y)
            pdf_doc.line(x1, y2, x2, y2)
            if self.table.border == Border.grid:
                # draw lines between table columns
                cells = self.bands[0]['cells']
                # add half border_width so border is drawn inside right cell and
                # can be aligned with borders of other elements outside the table
                x = x1
                y2 = y1 + self.bands[0]['height']

                # rows can have different cells (colspan) than other rows so
                # we draw cell borders separately if necessary
                for band in self.bands[1:]:
                    current_cells = band['cells']
                    same_borders = True
                    if len(cells) == len(current_cells):
                        for (col1, col2) in zip(cells, current_cells):
                            if col1.width != col2.width:
                                same_borders = False
                                break
                    else:
                        same_borders = False
                    if not same_borders:
                        x = x1
                        for cell in cells[:-1]:
                            x += cell.width
                            pdf_doc.line(x, y1, x, y2)
                        y1 = y2
                        x = x1
                        cells = current_cells
                    y2 += band['height']

                for cell in cells[:-1]:
                    x += cell.width
                    pdf_doc.line(x, y1, x, y2)

    def cleanup(self):
        for band in self.bands:
            for element in band['elements']:
                element.cleanup()


class FrameRenderElement(DocElementBase):
    def __init__(self, report, frame, render_y):
        DocElementBase.__init__(self, report, dict(y=0))
        self.report = report
        self.x = frame.x
        self.width = frame.width
        self.border_style = frame.border_style
        self.background_color = frame.background_color
        self.render_y = render_y
        self.render_bottom = render_y
        self.height = 0
        self.elements = []
        self.render_element_type = RenderElementType.none
        self.complete = False

    def add_elements(self, container, render_element_type, height):
        self.elements = list(container.render_elements)
        self.render_element_type = render_element_type
        self.render_bottom += height

    def render_pdf(self, container_offset_x, container_offset_y, pdf_doc):
        x = self.x + container_offset_x
        y = self.render_y + container_offset_y
        height = self.render_bottom - self.render_y

        content_x = x
        content_width = self.width
        content_y = y
        content_height = height

        if self.border_style.border_left:
            content_x += self.border_style.border_width
            content_width -= self.border_style.border_width
        if self.border_style.border_right:
            content_width -= self.border_style.border_width
        if self.border_style.border_top and\
                self.render_element_type in (RenderElementType.first, RenderElementType.complete):
            content_y += self.border_style.border_width
            content_height -= self.border_style.border_width
        if self.border_style.border_bottom and\
                self.render_element_type in (RenderElementType.last, RenderElementType.complete):
            content_height -= self.border_style.border_width

        if not self.background_color.transparent:
            pdf_doc.set_fill_color(self.background_color.r, self.background_color.g, self.background_color.b)
            pdf_doc.rect(content_x, content_y, content_width, content_height, style='F')

        render_y = y
        if self.border_style.border_top and\
                self.render_element_type in (RenderElementType.first, RenderElementType.complete):
            render_y += self.border_style.border_width
        for element in self.elements:
            element.render_pdf(container_offset_x=content_x, container_offset_y=content_y, pdf_doc=pdf_doc)

        if (self.border_style.border_left or self.border_style.border_top or
                self.border_style.border_right or self.border_style.border_bottom):
            DocElement.draw_border(
                x=x, y=y, width=self.width, height=height,
                render_element_type=self.render_element_type, border_style=self.border_style, pdf_doc=pdf_doc)

    def cleanup(self):
        for element in self.elements:
            element.cleanup()


class SectionRenderElement(DocElementBase):
    def __init__(self, report, render_y):
        DocElementBase.__init__(self, report, dict(y=0))
        self.report = report
        self.render_y = render_y
        self.render_bottom = render_y
        self.height = 0
        self.bands = []
        self.complete = False

    def is_empty(self):
        return len(self.bands) == 0

    def add_section_band(self, section_band):
        if section_band.rendering_complete or not section_band.always_print_on_same_page:
            band_height = section_band.get_used_band_height()
            self.bands.append(dict(height=band_height, elements=list(section_band.get_render_elements())))
            self.height += band_height
            self.render_bottom += band_height

    def render_pdf(self, container_offset_x, container_offset_y, pdf_doc):
        y = self.render_y + container_offset_y
        for band in self.bands:
            for element in band['elements']:
                element.render_pdf(container_offset_x=container_offset_x, container_offset_y=y, pdf_doc=pdf_doc)
            y += band['height']

    def cleanup(self):
        for band in self.bands:
            for element in band['elements']:
                element.cleanup()





class ReportBroError(Exception):
    def __init__(self, error):
        self.error = error

    def __str__(self):
        return 'ReportBroError: ' + ', '.join(
            ['{key}={value}'.format(key=key, value=self.error.get(key)) for key in self.error])


class Error(dict):
    def __init__(self, msg_key, object_id=None, field=None, info=None, context=None):
        dict.__init__(self, msg_key=msg_key, object_id=object_id, field=field, info=info, context=context)



from enum import Enum


class Border(Enum):
    grid = 1
    frame_row = 2
    frame = 3
    row = 4
    none = 5


class RenderElementType(Enum):
    none = 0
    complete = 1
    between = 2
    first = 3
    last = 4


class DocElementType(Enum):
    text = 1
    image = 2
    line = 3
    page_break = 4
    table_band = 5
    table = 6
    table_text = 7
    bar_code = 8
    frame = 9
    section = 10
    section_band = 11


class ParameterType(Enum):
    none = 0
    string = 1
    number = 2
    boolean = 3
    date = 4
    array = 5
    simple_array = 6
    map = 7
    sum = 8
    average = 9
    image = 10


class BandType(Enum):
    header = 1
    content = 2
    footer = 3


class PageFormat(Enum):
    a4 = 1
    a5 = 2
    letter = 3
    user_defined = 4


class Unit(Enum):
    pt = 1
    mm = 2
    inch = 3


class Orientation(Enum):
    portrait = 1
    landscape = 2


class BandDisplay(Enum):
    never = 1
    always = 2
    not_on_first_page = 3


class HorizontalAlignment(Enum):
    left = 1
    center = 2
    right = 3
    justify = 4


class VerticalAlignment(Enum):
    top = 1
    middle = 2
    bottom = 3



#
# Copyright (C) 2017-2020 jobsta
#
# This file is part of ReportBro, a library to generate PDF and Excel reports.
# Demos can be found at https://www.reportbro.com
#
# Dual licensed under AGPLv3 and ReportBro commercial license:
# https://www.reportbro.com/license
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see https://www.gnu.org/licenses/
#
# Details for ReportBro commercial license can be found at
# https://www.reportbro.com/license/agreement
#

from __future__ import unicode_literals
from __future__ import division
import base64
import fpdf
import re
import xlsxwriter
import pkg_resources
from io import BytesIO, BufferedReader
import os

from .containers import ReportBand
from .elements import *
from .enums import *
from .structs import Parameter, TextStyle
from .utils import get_int_value, parse_datetime_string, PY3


try:
    basestring  # For Python 2, str and unicode
except NameError:
    basestring = str

if PY3:
    long = int

regex_valid_identifier = re.compile(r'^[^\d\W]\w*$', re.U)


class DocumentPDFRenderer:
    def __init__(self, header_band, content_band, footer_band, report, context,
                 additional_fonts, filename, add_watermark, page_limit, encode_error_handling, core_fonts_encoding):
        self.header_band = header_band
        self.content_band = content_band
        self.footer_band = footer_band
        self.document_properties = report.document_properties
        self.pdf_doc = FPDFRB(
            report.document_properties, additional_fonts=additional_fonts,
            encode_error_handling=encode_error_handling, core_fonts_encoding=core_fonts_encoding)
        self.pdf_doc.set_margins(0, 0)
        self.pdf_doc.c_margin = 0  # interior cell margin
        self.context = context
        self.filename = filename
        self.add_watermark = add_watermark
        self.page_limit = page_limit

    def add_page(self):
        self.pdf_doc.add_page()
        self.context.inc_page_number()

    def is_finished(self):
        return self.content_band.is_finished()

    def render(self):
        watermark_width = watermark_height = 0
        watermark_filename = pkg_resources.resource_filename('reportbro', 'data/logo_watermark.png')
        if self.add_watermark:
            if not os.path.exists(watermark_filename):
                self.add_watermark = False
            else:
                watermark_width = self.document_properties.page_width / 3
                watermark_height = watermark_width * (115 / 460)

        self.content_band.prepare(self.context, self.pdf_doc)
        page_count = 1
        while True:
            height = self.document_properties.page_height -\
                self.document_properties.margin_top - self.document_properties.margin_bottom
            if self.document_properties.header_display == BandDisplay.always or\
                    (self.document_properties.header_display == BandDisplay.not_on_first_page and page_count != 1):
                height -= self.document_properties.header_size
            if self.document_properties.footer_display == BandDisplay.always or\
                    (self.document_properties.footer_display == BandDisplay.not_on_first_page and page_count != 1):
                height -= self.document_properties.footer_size
            complete = self.content_band.create_render_elements(0, height, self.context, self.pdf_doc)
            if complete:
                break
            page_count += 1
            if self.page_limit and page_count > self.page_limit:
                raise RuntimeError('Too many pages (probably an endless loop)')
        self.context.set_page_count(page_count)

        footer_offset_y = self.document_properties.page_height -\
            self.document_properties.footer_size - self.document_properties.margin_bottom
        # render at least one page to show header/footer even if content is empty
        while not self.content_band.is_finished() or self.context.get_page_number() == 0:
            self.add_page()
            if self.add_watermark:
                if watermark_height < self.document_properties.page_height:
                    self.pdf_doc.image(
                        watermark_filename,
                        self.document_properties.page_width / 2 - watermark_width / 2,
                        self.document_properties.page_height - watermark_height,
                        watermark_width, watermark_height)

            content_offset_y = self.document_properties.margin_top
            page_number = self.context.get_page_number()
            if self.document_properties.header_display == BandDisplay.always or\
                    (self.document_properties.header_display == BandDisplay.not_on_first_page and page_number != 1):
                content_offset_y += self.document_properties.header_size
                self.header_band.prepare(self.context, self.pdf_doc)
                self.header_band.create_render_elements(
                    0, self.document_properties.header_size, self.context, self.pdf_doc)
                self.header_band.render_pdf(
                    self.document_properties.margin_left, self.document_properties.margin_top, self.pdf_doc)
                self.header_band.reset()
            if self.document_properties.footer_display == BandDisplay.always or\
                    (self.document_properties.footer_display == BandDisplay.not_on_first_page and page_number != 1):
                self.footer_band.prepare(self.context, self.pdf_doc)
                self.footer_band.create_render_elements(
                    0, self.document_properties.footer_size, self.context, self.pdf_doc)
                self.footer_band.render_pdf(self.document_properties.margin_left, footer_offset_y, self.pdf_doc)
                self.footer_band.reset()

            self.content_band.render_pdf(
                self.document_properties.margin_left, content_offset_y, self.pdf_doc, cleanup=True)

        self.header_band.cleanup()
        self.footer_band.cleanup()
        dest = 'F' if self.filename else 'S'
        return self.pdf_doc.output(name=self.filename, dest=dest)


class DocumentXLSXRenderer:
    def __init__(self, header_band, content_band, footer_band, report, context, filename):
        self.header_band = header_band
        self.content_band = content_band
        self.footer_band = footer_band
        self.document_properties = report.document_properties
        self.workbook_mem = BytesIO()
        self.workbook = xlsxwriter.Workbook(filename if filename else self.workbook_mem)
        self.worksheet = self.workbook.add_worksheet()
        self.context = context
        self.filename = filename
        self.row = 0
        self.column_widths = []

    def render(self):
        if self.document_properties.header_display != BandDisplay.never:
            self.render_band(self.header_band)
        self.render_band(self.content_band)
        if self.document_properties.header_display != BandDisplay.never:
            self.render_band(self.footer_band)

        for i, column_width in enumerate(self.column_widths):
            if column_width > 0:
                # setting the column width is just an approximation, in Excel the width
                # is the number of characters in the default font
                self.worksheet.set_column(i, i, column_width / 7)

        self.workbook.close()
        if not self.filename:
            # if no filename is given the spreadsheet data will be returned
            self.workbook_mem.seek(0)
            return self.workbook_mem.read()
        return None

    def render_band(self, band):
        band.prepare(self.context)
        self.row, _ = band.render_spreadsheet(self.row, 0, self.context, self)

    def update_column_width(self, col, width):
        if col >= len(self.column_widths):
            # make sure column_width list contains entries for each column
            self.column_widths.extend([-1] * (col + 1 - len(self.column_widths)))
        if width > self.column_widths[col]:
            self.column_widths[col] = width

    def write(self, row, col, colspan, text, cell_format, width, url=None):
        if colspan > 1:
            self.worksheet.merge_range(row, col, row, col + colspan - 1, text, cell_format)
        elif not url:
            self.worksheet.write(row, col, text, cell_format)
            self.update_column_width(col, width)
        # url also works combined with colspan, the first cell of the range is simply overwritten
        if url:
            self.worksheet.write_url(row, col, url, cell_format, text)

    def insert_image(self, row, col, image_filename, image_data, width, url=None):
        options = dict()
        if image_data:
            options['image_data'] = image_data
        if url:
            options['url'] = url
        self.worksheet.insert_image(row, col, image_filename, options)
        self.update_column_width(col, width)

    def add_format(self, format_props):
        return self.workbook.add_format(format_props)

    def set_row(self, row, cell_format):
        self.worksheet.set_row(row, cell_format=cell_format)


class DocumentProperties:
    def __init__(self, report, data):
        self.id = '0_document_properties'
        self.page_format = PageFormat[data.get('pageFormat').lower()]
        self.orientation = Orientation[data.get('orientation')]
        self.report = report

        if self.page_format == PageFormat.a4:
            if self.orientation == Orientation.portrait:
                self.page_width = 210
                self.page_height = 297
            else:
                self.page_width = 297
                self.page_height = 210
            unit = Unit.mm
        elif self.page_format == PageFormat.a5:
            if self.orientation == Orientation.portrait:
                self.page_width = 148
                self.page_height = 210
            else:
                self.page_width = 210
                self.page_height = 148
            unit = Unit.mm
        elif self.page_format == PageFormat.letter:
            if self.orientation == Orientation.portrait:
                self.page_width = 8.5
                self.page_height = 11
            else:
                self.page_width = 11
                self.page_height = 8.5
            unit = Unit.inch
        else:
            self.page_width = get_int_value(data, 'pageWidth')
            self.page_height = get_int_value(data, 'pageHeight')
            unit = Unit[data.get('unit')]
            if unit == Unit.mm:
                if self.page_width < 30 or self.page_width >= 100000:
                    self.report.errors.append(Error('errorMsgInvalidPageSize', object_id=self.id, field='page'))
                elif self.page_height < 30 or self.page_height >= 100000:
                    self.report.errors.append(Error('errorMsgInvalidPageSize', object_id=self.id, field='page'))
            elif unit == Unit.inch:
                if self.page_width < 1 or self.page_width >= 1000:
                    self.report.errors.append(Error('errorMsgInvalidPageSize', object_id=self.id, field='page'))
                elif self.page_height < 1 or self.page_height >= 1000:
                    self.report.errors.append(Error('errorMsgInvalidPageSize', object_id=self.id, field='page'))
        dpi = 72
        if unit == Unit.mm:
            self.page_width = round((dpi * self.page_width) / 25.4)
            self.page_height = round((dpi * self.page_height) / 25.4)
        else:
            self.page_width = round(dpi * self.page_width)
            self.page_height = round(dpi * self.page_height)

        self.content_height = get_int_value(data, 'contentHeight')
        self.margin_left = get_int_value(data, 'marginLeft')
        self.margin_top = get_int_value(data, 'marginTop')
        self.margin_right = get_int_value(data, 'marginRight')
        self.margin_bottom = get_int_value(data, 'marginBottom')
        self.pattern_locale = data.get('patternLocale', '')
        self.pattern_currency_symbol = data.get('patternCurrencySymbol', '')
        if self.pattern_locale not in ('de', 'en', 'es', 'fr', 'it'):
            raise RuntimeError('invalid pattern_locale')

        self.header = bool(data.get('header'))
        if self.header:
            self.header_display = BandDisplay[data.get('headerDisplay')]
            self.header_size = get_int_value(data, 'headerSize')
        else:
            self.header_display = BandDisplay.never
            self.header_size = 0
        self.footer = bool(data.get('footer'))
        if self.footer:
            self.footer_display = BandDisplay[data.get('footerDisplay')]
            self.footer_size = get_int_value(data, 'footerSize')
        else:
            self.footer_display = BandDisplay.never
            self.footer_size = 0
        if self.content_height == 0:
            self.content_height = self.page_height - self.header_size - self.footer_size -\
                self.margin_top - self.margin_bottom


class ImageData:
    def __init__(self, ctx, image_id, source, image_file, is_test_data, headers):
        self.image_fp = None
        self.image_type = None
        image_uri = None  # can be either url or file path
        image_url = None
        image_path = None
        img_data_b64 = None
        if source:
            if Context.is_parameter_name(source):
                param_ref = ctx.get_parameter(Context.strip_parameter_name(source))
                if param_ref:
                    source_parameter = param_ref.parameter
                    if source_parameter.type == ParameterType.string:
                        image_uri, _ = ctx.get_parameter_data(param_ref)
                    elif source_parameter.type == ParameterType.image:
                        # image is available as base64 encoded or
                        # file object (only possible if report data is passed directly from python code
                        # and not via web request)
                        img_data, _ = ctx.get_parameter_data(param_ref)
                        if isinstance(img_data, BufferedReader) or\
                                (PY2 and isinstance(img_data, file)):
                            self.image_fp = img_data
                            pos = img_data.name.rfind('.')
                            self.image_type = img_data.name[pos+1:].lower() if pos != -1 else ''
                        elif isinstance(img_data, basestring):
                            img_data_b64 = img_data
                    else:
                        raise ReportBroError(
                            Error('errorMsgInvalidImageSourceParameter', object_id=image_id, field='source'))
                else:
                    raise ReportBroError(
                        Error('errorMsgMissingParameter', object_id=image_id, field='source'))
            else:
                image_uri = source

        if img_data_b64 is None and not image_uri and self.image_fp is None and image_file:
            # static image base64 encoded within image element
            img_data_b64 = image_file

        if img_data_b64:
            m = re.match('^data:image/(.+);base64,', img_data_b64)
            if not m:
                raise ReportBroError(
                    Error('errorMsgInvalidImage', object_id=image_id, field='source'))
            self.image_type = m.group(1).lower()
            img_data = base64.b64decode(re.sub('^data:image/.+;base64,', '', img_data_b64))
            self.image_fp = BytesIO(img_data)
        elif image_uri:
            if image_uri.startswith("http://") or image_uri.startswith("https://"):
                image_url = image_uri
                try:
                    if PY2:
                        parse_result = urllib2.urlparse.urlparse(image_url)
                    else:
                        parse_result = urllib.parse.urlparse(image_url)
                    pos = parse_result.path.rfind('.')
                    self.image_type = parse_result.path[pos+1:].lower() if pos != -1 else ''
                except ValueError as ex:
                    raise ReportBroError(
                        Error('errorMsgInvalidImageSource', object_id=image_id, field='source'))
            elif not is_test_data and image_uri.startswith("file:"):
                # only allow image path (referencing file on server) when data is passed directly
                # and not from Reportbro Designer
                image_path = image_uri[5:]
                pos = image_uri.rfind('.')
                self.image_type = image_uri[pos+1:].lower() if pos != -1 else ''
            else:
                raise ReportBroError(
                    Error('errorMsgInvalidImageSource', object_id=image_id, field='source'))

        if self.image_type is not None:
            if self.image_type not in ('png', 'jpg', 'jpeg'):
                raise ReportBroError(
                    Error('errorMsgUnsupportedImageType', object_id=image_id, field='source'))

        if image_url:
            try:
                if PY2:
                    req = urllib2.Request(image_url, headers=headers)
                    self.image_fp = BytesIO(urllib2.urlopen(req).read())
                else:
                    req = urllib.request.Request(image_url, headers=headers)
                    self.image_fp = BytesIO(urllib.request.urlopen(req).read())
            except Exception as ex:
                raise ReportBroError(
                    Error('errorMsgLoadingImageFailed', object_id=image_id, field='source', info=str(ex)))
        elif image_path:
            try:
                cwd = os.getcwd()
                image_path = os.path.abspath(image_path)
                # make sure image file access is restricted to application
                if os.path.commonprefix([cwd, image_path]) != cwd:
                    raise Exception('Accessing file outside of application path not allowed')
                self.image_fp = open(image_path, 'rb')
            except Exception as ex:
                raise ReportBroError(
                    Error('errorMsgLoadingImageFailed', object_id=image_id, field='source', info=str(ex)))


class FPDFRB(fpdf.FPDF):
    def __init__(self, document_properties, additional_fonts, encode_error_handling, core_fonts_encoding):
        if document_properties.orientation == Orientation.portrait:
            orientation = 'P'
            dimension = (document_properties.page_width, document_properties.page_height)
        else:
            orientation = 'L'
            dimension = (document_properties.page_height, document_properties.page_width)
        fpdf.FPDF.__init__(self, orientation=orientation, unit='pt', format=dimension)
        self.x = 0
        self.y = 0
        self.set_doc_option('core_fonts_encoding', core_fonts_encoding)
        self.set_doc_option('encode_error_handling', encode_error_handling)
        self.loaded_images = dict()
        self.available_fonts = dict(
            courier=dict(standard_font=True),
            helvetica=dict(standard_font=True),
            times=dict(standard_font=True))
        if additional_fonts:
            for additional_font in additional_fonts:
                filename = additional_font.get('filename', '')
                font = dict(
                    standard_font=False, uni=additional_font.get('uni', True))

                regular_style = dict(
                    font_filename=filename, style='', font_added=False)
                bold_style = dict(
                    font_filename=additional_font.get('bold_filename', filename),
                    style='B', font_added=False)
                italic_style = dict(
                    font_filename=additional_font.get('italic_filename', filename),
                    style='I', font_added=False)
                bold_italic_style = dict(
                    font_filename=additional_font.get('bold_italic_filename', filename),
                    style='BI', font_added=False)

                # map styles in case there are no separate font-files for bold, italic or bold italic
                # to avoid adding the same font multiple times to the pdf document
                if bold_style['font_filename'] == regular_style['font_filename']:
                    bold_style = regular_style
                if italic_style['font_filename'] == regular_style['font_filename']:
                    italic_style = regular_style
                if bold_italic_style['font_filename'] == italic_style['font_filename']:
                    bold_italic_style = italic_style
                elif bold_italic_style['font_filename'] == bold_style['font_filename']:
                    bold_italic_style = bold_style
                elif bold_italic_style['font_filename'] == regular_style['font_filename']:
                    bold_italic_style = regular_style
                font['style'] = regular_style
                font['styleB'] = bold_style
                font['styleI'] = italic_style
                font['styleBI'] = bold_italic_style

                self.available_fonts[additional_font.get('value', '')] = font

    def add_image(self, img, image_key):
        self.loaded_images[image_key] = img

    def get_image(self, image_key):
        return self.loaded_images.get(image_key)

    def set_font(self, family, style='', size=0, underline=False):
        """Set font in underlying pdf renderer.

        This font is used for all following text rendering calls until changed again.

        :param family: name of the font which must either be one of the standard
        fonts (courier, helvetica, times) or added to ReportBro instance with the
        additional_fonts setting.
        :param size: font size, if not set then the standard font size is used.
        :param underline: True if text should be rendered with underlined style.
        :return: True if font exists, False otherwise.
        """
        font = self.available_fonts.get(family)
        if font:
            if not font['standard_font']:
                # get font for specific style
                if style:
                    # replace of 'U' is needed because it is set for underlined text
                    # when called from FPDF.add_page
                    style_font = font['style' + style.replace('U', '')]
                    # style could be different in case styles are mapped,
                    # e.g. if bold style has same font file as regular style
                    style = style_font['style']
                else:
                    style_font = font['style']

                if not style_font['font_added']:
                    self.add_font(
                        family, style=style, fname=style_font['font_filename'], uni=font['uni'])
                    style_font['font_added'] = True

            if underline:
                style += 'U'
            fpdf.FPDF.set_font(self, family, style, size)
            return True
        else:
            return False


class Report:
    def __init__(self, report_definition, data, is_test_data=False, additional_fonts=None,
                 page_limit=10000, request_headers=None, encode_error_handling='strict',
                 core_fonts_encoding='windows-1252'):
        """Create Report instance which can then be used to generate pdf and xlsx reports.

        :param report_definition: The report object containg report elements, parameters,
        styles and document properties. This object can be obtained in
        ReportBro Designer by using getReport method.
        :param data: Dictionary containing all data for the report.
        This structure must correspond with the defined parameters
        in the report_definition (parameter name and type).
        :param is_test_data: set to True in case the given data contains test data which
        is specified within the parameters in ReportBro Designer. Set to False if
        the data comes from your web application. This setting influences the
        error messages in case report generation fails due to invalid data.
        :param additional_fonts: In case additional (non-standard) fonts are used they
        must be made available so they can be embedded into the pdf file.
        :param page_limit: maximum number of pages for pdf reports. This can
        be used to avoid reports getting too big or taking too long for generation.
        If set to 0 or None then no page limit is used.
        :param request_headers: request headers used when images are fetched by url
        :param encode_error_handling: defines behaviour when a character cannot
        be encoded with the encoding used for the core fonts. The following options exist:
        - 'strict': raise a UnicodeDecodeError exception
        - 'ignore': just leave the character out of the result
        - 'replace': use U+FFFD replacement character
        :param core_fonts_encoding: defines the encoding when using the core fonts.
        Default is 'windows-1252' which is usually the best choice for English and many European
        languages including Spanish, French, and German.
        """
        assert isinstance(report_definition, dict)
        assert isinstance(data, dict)
        assert encode_error_handling in ('strict', 'ignore', 'replace')

        self.errors = []

        self.document_properties = DocumentProperties(self, report_definition.get('documentProperties'))

        self.containers = dict()
        self.header = ReportBand(BandType.header, '0_header', self.containers, self)
        self.content = ReportBand(BandType.content, '0_content', self.containers, self)
        self.footer = ReportBand(BandType.footer, '0_footer', self.containers, self)

        self.parameters = dict()
        self.styles = dict()
        self.data = dict()
        self.is_test_data = is_test_data

        self.additional_fonts = additional_fonts
        self.page_limit = page_limit
        self.encode_error_handling = encode_error_handling
        self.core_fonts_encoding = core_fonts_encoding
        # request headers used when fetching images by url (some sites check for existance
        # of user-agent header and do not return image otherwise)
        self.request_headers = {'User-Agent': 'Mozilla/5.0'}
        if request_headers is not None:
            self.request_headers = request_headers

        version = report_definition.get('version')
        if isinstance(version, int):
            # convert old report definitions
            if version < 2:
                for doc_element in report_definition.get('docElements'):
                    if DocElementType[doc_element.get('elementType')] == DocElementType.table:
                        doc_element['contentDataRows'] = [doc_element.get('contentData')]

        # list is needed to compute parameters (parameters with expression) in given order
        parameter_list = []
        for item in report_definition.get('parameters'):
            parameter = Parameter(self, item)
            if parameter.name in self.parameters:
                self.errors.append(Error('errorMsgDuplicateParameter', object_id=parameter.id, field='name'))
            self.parameters[parameter.name] = parameter
            parameter_list.append(parameter)

        for item in report_definition.get('styles'):
            style = TextStyle(item)
            style_id = int(item.get('id'))
            self.styles[style_id] = style

        for doc_element in report_definition.get('docElements'):
            element_type = DocElementType[doc_element.get('elementType')]
            container_id = str(doc_element.get('containerId'))
            container = None
            if container_id:
                container = self.containers.get(container_id)
            elem = None
            if element_type == DocElementType.text:
                elem = TextElement(self, doc_element)
            elif element_type == DocElementType.line:
                elem = LineElement(self, doc_element)
            elif element_type == DocElementType.image:
                elem = ImageElement(self, doc_element)
            elif element_type == DocElementType.bar_code:
                elem = BarCodeElement(self, doc_element)
            elif element_type == DocElementType.table:
                elem = TableElement(self, doc_element)
            elif element_type == DocElementType.page_break:
                elem = PageBreakElement(self, doc_element)
            elif element_type == DocElementType.frame:
                elem = FrameElement(self, doc_element, self.containers)
            elif element_type == DocElementType.section:
                elem = SectionElement(self, doc_element, self.containers)

            if elem and container:
                if container.is_visible():
                    if elem.x < 0:
                        self.errors.append(Error('errorMsgInvalidPosition', object_id=elem.id, field='x'))
                    elif elem.x + elem.width > container.width:
                        self.errors.append(Error('errorMsgInvalidSize', object_id=elem.id, field='width'))
                    if elem.y < 0:
                        self.errors.append(Error('errorMsgInvalidPosition', object_id=elem.id, field='y'))
                    elif elem.y + elem.height > container.height:
                        self.errors.append(Error('errorMsgInvalidSize', object_id=elem.id, field='height'))
                container.add(elem)

        self.context = Context(self, self.parameters, self.data)

        self.images = dict()  # cached image data

        self.process_data(dest_data=self.data, src_data=data, parameters=parameter_list,
                          is_test_data=is_test_data, parents=[])
        try:
            if not self.errors:
                self.evaluate_parameters(parameter_list, self.data)
        except ReportBroError as err:
            self.errors.append(err.error)

    def load_image(self, image_key, ctx, image_id, source, image_file):
        # test if image is not already loaded into image cache
        if image_key not in self.images:
            image = ImageData(ctx, image_id, source, image_file, self.is_test_data, headers=self.request_headers)
            self.images[image_key] = image

    def get_image(self, image_key):
        return self.images.get(image_key)

    def generate_pdf(self, filename='', add_watermark=False):
        renderer = DocumentPDFRenderer(
            header_band=self.header, content_band=self.content, footer_band=self.footer,
            report=self, context=self.context, additional_fonts=self.additional_fonts,
            filename=filename, add_watermark=add_watermark, page_limit=self.page_limit,
            encode_error_handling=self.encode_error_handling, core_fonts_encoding=self.core_fonts_encoding)
        return renderer.render()

    def generate_xlsx(self, filename=''):
        renderer = DocumentXLSXRenderer(header_band=self.header, content_band=self.content, footer_band=self.footer,
                report=self, context=self.context, filename=filename)
        return renderer.render()

    # goes through all elements in header, content and footer and throws a ReportBroError in case
    # an element is invalid
    def verify(self):
        if self.document_properties.header_display != BandDisplay.never:
            self.header.prepare(self.context, only_verify=True)
        self.content.prepare(self.context, only_verify=True)
        if self.document_properties.header_display != BandDisplay.never:
            self.footer.prepare(self.context, only_verify=True)

    def parse_parameter_value(self, parameter, parent_id, is_test_data, parameter_type, value):
        error_field = 'test_data' if is_test_data else 'type'
        if parameter_type == ParameterType.string:
            if value is not None:
                if not isinstance(value, basestring):
                    raise RuntimeError('value of parameter {name} must be str type (unicode for Python 2.7.x)'.
                            format(name=parameter.name))
            elif not parameter.nullable:
                value = ''

        elif parameter_type == ParameterType.number:
            if value:
                if isinstance(value, basestring):
                    value = value.replace(',', '.')
                try:
                    value = decimal.Decimal(str(value))
                except (decimal.InvalidOperation, TypeError):
                    if parent_id and is_test_data:
                        self.errors.append(Error('errorMsgInvalidTestData', object_id=parent_id, field='test_data'))
                        self.errors.append(Error('errorMsgInvalidNumber', object_id=parameter.id, field='type'))
                    else:
                        self.errors.append(Error('errorMsgInvalidNumber',
                                                 object_id=parameter.id, field=error_field, context=parameter.name))
            elif value is not None:
                if isinstance(value, (int, long, float)):
                    value = decimal.Decimal(0)
                elif is_test_data and isinstance(value, basestring):
                    value = None if parameter.nullable else decimal.Decimal(0)
                elif not isinstance(value, decimal.Decimal):
                    if parent_id and is_test_data:
                        self.errors.append(Error('errorMsgInvalidTestData', object_id=parent_id, field='test_data'))
                        self.errors.append(Error('errorMsgInvalidNumber', object_id=parameter.id, field='type'))
                    else:
                        self.errors.append(Error('errorMsgInvalidNumber',
                                                 object_id=parameter.id, field=error_field, context=parameter.name))
            elif not parameter.nullable:
                value = decimal.Decimal(0)

        elif parameter_type == ParameterType.boolean:
            if value is not None:
                value = bool(value)
            elif not parameter.nullable:
                value = False

        elif parameter_type == ParameterType.date:
            if isinstance(value, basestring):
                if is_test_data and not value:
                    value = None if parameter.nullable else datetime.datetime.now()
                else:
                    try:
                        value = parse_datetime_string(value)
                    except (ValueError, TypeError):
                        if parent_id and is_test_data:
                            self.errors.append(Error('errorMsgInvalidTestData', object_id=parent_id, field='test_data'))
                            self.errors.append(Error('errorMsgInvalidDate', object_id=parameter.id, field='type'))
                        else:
                            self.errors.append(Error('errorMsgInvalidDate',
                                                     object_id=parameter.id, field=error_field, context=parameter.name))
            elif isinstance(value, datetime.date):
                if not isinstance(value, datetime.datetime):
                    value = datetime.datetime(value.year, value.month, value.day)
            elif value is not None:
                if parent_id and is_test_data:
                    self.errors.append(Error('errorMsgInvalidTestData', object_id=parent_id, field='test_data'))
                    self.errors.append(Error('errorMsgInvalidDate', object_id=parameter.id, field='type'))
                else:
                    self.errors.append(Error('errorMsgInvalidDate',
                                             object_id=parameter.id, field=error_field, context=parameter.name))
            elif not parameter.nullable:
                value = datetime.datetime.now()
        return value

    def process_data(self, dest_data, src_data, parameters, is_test_data, parents):
        field = 'test_data' if is_test_data else 'type'
        parent_id = parents[-1].id if parents else None
        for parameter in parameters:
            if parameter.is_internal:
                continue
            if regex_valid_identifier.match(parameter.name) is None:
                self.errors.append(Error('errorMsgInvalidParameterName',
                                         object_id=parameter.id, field='name', info=parameter.name))
            parameter_type = parameter.type
            if not parameter.is_evaluated():
                value = src_data.get(parameter.name)
                if parameter_type in (ParameterType.string, ParameterType.number,
                                      ParameterType.boolean, ParameterType.date):
                    value = self.parse_parameter_value(parameter, parent_id, is_test_data, parameter_type, value)
                elif not parents:
                    if parameter_type == ParameterType.array:
                        if isinstance(value, list):
                            parents.append(parameter)
                            parameter_list = list(parameter.fields.values())
                            # create new list which will be assigned to dest_data to keep src_data unmodified
                            dest_array = []

                            for row_number, row in enumerate(value, start=1):
                                dest_array_item = dict()
                                self.process_data(
                                    dest_data=dest_array_item, src_data=row, parameters=parameter_list,
                                    is_test_data=is_test_data, parents=parents)
                                # set value for internal parameter 'row_number'
                                dest_array_item['row_number'] = row_number
                                dest_array.append(dest_array_item)
                            parents = parents[:-1]
                            value = dest_array
                        elif value is None:
                            if not parameter.nullable:
                                value = []
                        else:
                            self.errors.append(Error(
                                'errorMsgInvalidArray',
                                object_id=parameter.id, field=field, context=parameter.name))
                    elif parameter_type == ParameterType.simple_array:
                        if isinstance(value, list):
                            list_values = []
                            for list_value in value:
                                parsed_value = self.parse_parameter_value(
                                    parameter, parent_id, is_test_data, parameter.array_item_type, list_value)
                                list_values.append(parsed_value)
                            value = list_values
                        elif value is None:
                            if not parameter.nullable:
                                value = []
                        else:
                            self.errors.append(Error(
                                'errorMsgInvalidArray',
                                object_id=parameter.id, field=field, context=parameter.name))
                    elif parameter_type == ParameterType.map:
                        if value is None and not parameter.nullable:
                            value = dict()
                        if isinstance(value, dict):
                            if isinstance(parameter.children, list):
                                parents.append(parameter)
                                # create new dict which will be assigned to dest_data to keep src_data unmodified
                                dest_map = dict()

                                self.process_data(
                                    dest_data=dest_map, src_data=value, parameters=parameter.children,
                                    is_test_data=is_test_data, parents=parents)
                                parents = parents[:-1]
                                value = dest_map
                            else:
                                self.errors.append(Error(
                                    'errorMsgInvalidMap',
                                    object_id=parameter.id, field='type', context=parameter.name))
                        else:
                            self.errors.append(Error(
                                'errorMsgMissingData',
                                object_id=parameter.id, field='name', context=parameter.name))
                dest_data[parameter.name] = value

    def evaluate_parameters(self, parameters, data):
        for parameter in parameters:
            if not parameter.is_internal:
                if parameter.is_evaluated():
                    self.evaluate_parameter_expr(parameter, data)
                elif parameter.type == ParameterType.map:
                    for field in parameter.children:
                        if field.is_evaluated():
                            # set dest_data so evaluated expression is set in map
                            self.evaluate_parameter_expr(field, data, dest_data=data[parameter.name])
                elif parameter.type == ParameterType.array:
                    eval_fields = []
                    for field in parameter.children:
                        if field.eval:
                            eval_fields.append(field)
                    if eval_fields:
                        param_ref = self.context.get_parameter(parameter.name)
                        if param_ref is not None:
                            rows, data_exists = self.context.get_parameter_data(param_ref)
                            if data_exists:
                                row_parameters = dict()
                                for row_parameter in parameter.children:
                                    row_parameters[row_parameter.name] = row_parameter

                                for row in rows:
                                    self.context.push_context(row_parameters, row)
                                    for field in eval_fields:
                                        self.evaluate_parameter_expr(field, row)
                                    self.context.pop_context()

    def evaluate_parameter_expr(self, parameter, data, dest_data=None):
        if not parameter.expression:
            self.errors.append(Error(
                'errorMsgMissingExpression',
                object_id=parameter.id, field='expression', context=parameter.name))
            return

        parameter_type = parameter.type
        if parameter_type in (ParameterType.average, ParameterType.sum):
            value, valid_value = self.context.evaluate_parameter_func(parameter)
        else:
            value = self.context.evaluate_expression(
                parameter.expression, parameter.id, field='expression')
            valid_value = False
            if parameter_type == ParameterType.string:
                if isinstance(value, basestring):
                    valid_value = True
            elif parameter_type == ParameterType.number:
                if isinstance(value, decimal.Decimal):
                    valid_value = True
                elif isinstance(value, (int, long, float)):
                    value = decimal.Decimal(value)
                    valid_value = True
            elif parameter_type == ParameterType.boolean:
                if isinstance(value, bool):
                    valid_value = True
            elif parameter_type == ParameterType.date:
                if isinstance(value, basestring):
                    try:
                        value = parse_datetime_string(value)
                    except (ValueError, TypeError):
                        self.errors.append(Error(
                            'errorMsgInvalidExpressionType',
                            object_id=parameter.id, field='expression', context=parameter.name))
                    valid_value = True
                elif isinstance(value, datetime.date):
                    valid_value = True
                    if not isinstance(value, datetime.datetime):
                        value = datetime.datetime(value.year, value.month, value.day)

            if not valid_value:
                self.errors.append(Error(
                    'errorMsgInvalidExpressionType',
                    object_id=parameter.id, field='expression', context=parameter.name))

        if valid_value:
            if dest_data is not None:
                dest_data[parameter.name] = value
            else:
                data[parameter.name] = value
